C    Odeidon, a program for structural identification of ODEs
C    Copyright (C) 2011 Peter Gennemark, Dag Wedelin
C
C    This program is free software: you can redistribute it and/or modify
C    it under the terms of the GNU General Public License as published by
C    the Free Software Foundation, either version 3 of the License, or
C    (at your option) any later version.
C
C    This program is distributed in the hope that it will be useful,
C    but WITHOUT ANY WARRANTY; without even the implied warranty of
C    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C    GNU General Public License for more details.

C    You should have received a copy of the GNU General Public License
C    along with this program.  If not, see <http://www.gnu.org/licenses/>.

C    Contact: Peter Gennemark, Dag Wedelin
C             www.odeidentification.org
C
C
C
C     GLOBAL SEARCH WITH SIM OF COMPLETE MODEL
C     COMPUTES THE RESIDUAL VECTOR, R = R(X)
C
C     N - number of data-points (IN)
C     P - number of parameters (IN)
C     X - parameters, length P (IN)
C     NF - invocation count, see below  (IN/OUT)
C     R - residuals (OUT)
C     LTY - additional integer input, here doesn't used (IN)
C     TY - additional real input, here doesn't used (IN)
C     UF - additional function, here doesn't used (IN)
C
C     WHEN CALCR IS CALLED, NF IS THE INVOCATION COUNT
C     FOR CALCR.  IT IS INCLUDED FOR POSSIBLE USE WITH
C     CALCJ.  IF X IS OUT OF BOUNDS (E.G. IF IT WOULD
C     CAUSE OVERFLOW IN COMPUTING R(X)), THEN CALCR SHOULD
C     SET NF TO 0.  THIS WILL CAUSE A SHORTER STEP TO BE
C     ATTEMPTED.   
C
C     DO NOT CHANGE N, P, OR X.
C
      SUBROUTINE FUN1(N, P, X, NF, R, LTY, TY, UF) 

      implicit none    
      INTEGER N, P, NF, LTY 
      DOUBLE PRECISION X(P), R(N), TY(1) 
      EXTERNAL UF 
      integer nvar,nexp
      integer maxexp,maxvar,maxpts,maxr,maxv,maxk
      parameter (maxexp=20,maxvar=19,maxpts=201,maxr=10,maxv=3,maxk=5)


      
C     ------ MODEL ------
      integer modtype(maxvar,maxr) 
C     Type of reaction [1...maxr] for variable [1...maxvar]
      double precision modk(maxvar,maxr,maxk,3) 
C     Last index=1: Parameter [1...maxk] for reaction [1...maxr] in variable [1...maxvar]
C     Last index=2: Lower bound for parameter [1...maxk] for reaction [1...maxr] in variable [1...maxvar]
C     Last index=3: Upper bound for parameter [1...maxk] for reaction [1...maxr] in variable [1...maxvar]
      double precision modkinit(maxexp,maxvar,3) 
C     Last index=1: Init conc for experiment [1...maxexp], variable [1...maxvar]
C     Last index=2: Lower bound for experiment [1...maxexp], variable [1...maxvar]
C     Last index=3: Upper bound for experiment [1...maxexp], variable [1...maxvar]
      integer modn(maxvar) 
C     Number of reactions for variable [1...maxvar]
      integer modv(maxvar,maxr,maxv)
C     Variable [1...maxv] for reaction [1...maxr] in variable [1...maxvar]
      logical dependent(maxvar) 
C     Which variables are dependent

      integer nvare,npts(maxexp),npts2
      double precision edata(maxexp,maxvar+1,2*maxpts)
      integer spline(maxexp,maxvar) ! 0=linear interpolation, 1=cubic smoothing spline 

      double precision RDUM(N) ! residuals
      double precision error

      !logical initconc          ! T if init conc consider as parameter, ow F
      integer debug,i,jj,ieq
      DOUBLE PRECISION B(2,400)

      common/mydata/edata,spline,npts,nexp,nvar,nvare
      common/mydata3/modk,modkinit,modn,modv,modtype,dependent
      common/myBest/B ! communicates est-fun2


      do 1 jj=1,P
         if (X(jj).ge.B(1,jj).and.X(jj).le.B(2,jj)) then
         else
            !write(*,*) 'fun1, x out of bounds ', X(jj)
            do 630 ieq=1,N
               R(ieq)=1.d10
 630        enddo
            NF=0
            return
         endif
 1    enddo


      debug=0
      if (debug.gt.0) then 
         write(*,*) 'FUN1',(X(i),i=1,P)
      endif   
      
      call xtomodel (maxvar,maxr,maxv,maxk,nvar,
     *     modtype,modk,modkinit,modn,modv,dependent,maxexp,nexp,X,P)

      
      call errtru(error,RDUM,R,N)


      if (debug.gt.0.and.error.gt.1.d100) then 
         write(*,*) 'END FUN1',(X(i),i=1,P)
         write(*,*) 'eval FUN1, error ',error
         write(*,*) 'R (1..21): ',(R(i),i=1,21)
         call  printmodel(maxvar,maxr,maxv,maxk,nvar,
     *        modtype,modk,modn,modv,dependent)
         stop
      endif
      RETURN 
      END 





C     GLOBAL SEARCH WITH SIM OF COMPLETE MODEL
C     COMPUTES THE JACOBIAN MATRIX, J = J(X)
C     J(I,K) IS SET TO THE PARTIAL 
C     DERIVATIVE OF COMPONENT I OF R WITH RESPECT TO X(K).
C     
C     NOTE: LTY the variable to consider
C
C     N - number of data-points (IN)
C     P - number of parameters (IN)
C     X - parameters, length P (IN)
C     NF - invocation count, see below  (IN/OUT)
C     J - Jacobian (OUT)
C     LTY - additional integer input, here the variable to consider (IN)
C     TY - additional real input, here doesn't used (IN)
C     UF - additional function, here doesn't used (IN)
C
C     A SUBROUTINE WHICH, GIVEN X, COMPUTES THE JACOBIAN
C     MATRIX J OF R AT X, I.E., THE N BY P MATRIX WHOSE
C     (I,K) ENTRY IS THE PARTIAL DERIVATIVE OF THE I-TH
C     COMPONENT OF R WITH RESPECT TO X(K).  CALCJ MUST BE
C     DECLARED EXTERNAL IN THE CALLING PROGRAM.  IT IS
C     INVOKED BY
C     CALL CALCJ(N,P,X,NF,J,UIPARM,URPARM,UFPARM)
C     NF IS THE INVOCATION COUNT FOR CALCR AT THE TIME
C     R(X) WAS EVALUATED.  THE X PASSED TO CALCJ IS
C     USUALLY THE ONE PASSED TO CALCR ON EITHER ITS MOST
C     RECENT INVOCATION OR THE ONE PRIOR TO IT.  IF CALCR
C     SAVES INTERMEDIATE RESULTS FOR USE BY CALCJ, THEN IT
C     IS POSSIBLE TO TELL FROM NF WHETHER THEY ARE VALID
C     FOR THE CURRENT X (OR WHICH COPY IS VALID IF TWO
C     COPIES ARE KEPT).  IF J CANNOT BE COMPUTED AT X,
C     THEN CALCJ SHOULD SET NF TO 0.  IN THIS CASE, NL2SOL
C     WILL RETURN WITH IV(1) = 15.  THE OTHER PARAMETERS
C     TO CALCJ ARE AS DESCRIBED ABOVE AND BELOW.  
C
C     DO NOT CHANGE N, P, OR X.
C
      SUBROUTINE JACOB1(N, P, X, NF, J, LTY, TY, UF) 
      implicit none
      INTEGER N, P, NF, LTY 
      DOUBLE PRECISION X(P), J(N,P), TY(1) 
      EXTERNAL UF 
      INTEGER I,IEQ     

      integer nvar,nexp
      integer maxexp,maxvar,maxpts,maxr,maxv,maxk
      parameter (maxexp=20,maxvar=19,maxpts=201,maxr=10,maxv=3,maxk=5)

      
C     ------ MODEL ------
      integer modtype(maxvar,maxr) 
C     Type of reaction [1...maxr] for variable [1...maxvar]
      double precision modk(maxvar,maxr,maxk,3) 
C     Last index=1: Parameter [1...maxk] for reaction [1...maxr] in variable [1...maxvar]
C     Last index=2: Lower bound for parameter [1...maxk] for reaction [1...maxr] in variable [1...maxvar]
C     Last index=3: Upper bound for parameter [1...maxk] for reaction [1...maxr] in variable [1...maxvar]
      double precision modkinit(maxexp,maxvar,3) 
C     Last index=1: Init conc for experiment [1...maxexp], variable [1...maxvar]
C     Last index=2: Lower bound for experiment [1...maxexp], variable [1...maxvar]
C     Last index=3: Upper bound for experiment [1...maxexp], variable [1...maxvar]
      integer modn(maxvar) 
C     Number of reactions for variable [1...maxvar]
      integer modv(maxvar,maxr,maxv)
C     Variable [1...maxv] for reaction [1...maxr] in variable [1...maxvar]
      logical dependent(maxvar) 
C     Which variables are dependent

      integer nvare,npts(maxexp),kn,iexp,ivar,ir
      double precision edata(maxexp,maxvar+1,2*maxpts)
      integer spline(maxexp,maxvar) ! 0=linear interpolation, 1=cubic smoothing spline 

      double precision R(N),RORG(N),RES(N) ! residuals
      double precision error,DELTA

      logical initconc          ! T if init conc consider as parameter, ow F
      integer debug

      common/mydata/edata,spline,npts,nexp,nvar,nvare
      common/mydata3/modk,modkinit,modn,modv,modtype,dependent
      common/myinitconc/initconc
      
      debug=0

      if (debug.gt.0) then 
         write(*,*) 'JAC1',(X(i),i=1,P)
      endif
      call xtomodel (maxvar,maxr,maxv,maxk,nvar,
     *     modtype,modk,modkinit,modn,modv,dependent,maxexp,nexp,X,P)
      call errtru(error,RORG,RES,N)
      
      if (debug.gt.1) then  
         write(*,*) 'GET JACOBICAN'
      endif      

      
      call getKn(maxvar,maxr,maxv,maxk,nvar,
     *     modtype,modk,modn,modv,dependent,kn)
      
   
      do 20 I=1,kn
         DELTA=ABS(X(I))*1.d-6
C     This following is needed to avoid division by zero further down
C     when X(I) is close to zero
         DELTA=MAX(DELTA,1.d-14) 
         X(I)=X(I)+DELTA        ! DELTA CHANGE
         call xtomodel (maxvar,maxr,maxv,maxk,nvar,
     *        modtype,modk,modkinit,modn,modv,dependent,maxexp,nexp,X,P)      
         call errtru(error,R,RES,N)
         X(I)=X(I)-DELTA        !RESET
         call xtomodel (maxvar,maxr,maxv,maxk,nvar,
     *        modtype,modk,modkinit,modn,modv,dependent,maxexp,nexp,X,P) 
         do 10 IEQ=1,N
            J(IEQ,I)=-(R(IEQ)-RORG(IEQ))/DELTA
 10      CONTINUE
 20   CONTINUE     
      

      I=kn+1
      if (initconc) then
         do 300 iexp=1,nexp
            do 200 ivar=1,nvar
               if (dependent(ivar)) then
                  DELTA=ABS(X(I))*1.d-6
C     This following is needed to avoid division by zero further down
C     when X(I) is close to zero
                  DELTA=MAX(DELTA,1.d-14) 
                  X(I)=X(I)+DELTA ! DELTA CHANGE
                  call xtomodel (maxvar,maxr,maxv,maxk,nvar,
     *                 modtype,modk,modkinit,modn,modv,dependent,
     *                 maxexp,nexp,X,P)
                  do 60 ir=1,N
                     R(ir)=RORG(ir) ! Reset R 
 60               enddo
                  call errtruforexp(error,R,RES,N,iexp) ! update part of R
                  X(I)=X(I)-DELTA !RESET
                  call xtomodel (maxvar,maxr,maxv,maxk,nvar,
     *                 modtype,modk,modkinit,modn,modv,dependent,
     *                 maxexp,nexp,X,P)
                  do 100 IEQ=1,N
                     J(IEQ,I)=-(R(IEQ)-RORG(IEQ))/DELTA
 100               CONTINUE
                  I=I+1
               endif
 200        continue
 300     continue
      endif


      if (I-1.ne.P) then
         write(*,*) 'JACOB1, I=',I,' P=',P,' kn',kn,initconc
         
         call  printmodel(maxvar,maxr,maxv,maxk,nvar,
     *        modtype,modk,modn,modv,dependent)
         stop
      endif



      
      if (debug.gt.0) then 
         write(*,*) 'eval JACOB1, error ',error
         if (debug.gt.1) then 
            write(*,*) 'R (1..11): ',(R(i),i=1,11)
            write(*,*) 'RES (1..11): ',(RES(i),i=1,11)
            write(*,*) 'J (1..11,1): ',(J(i,1),i=1,11)
         endif
      endif
      RETURN 
      END 


  









       
      

  



