% This Matlab script exemplifies how an ODEion problem file can be
% generated for a known system. The script can easily be modified, e.g., 
% to vary the noise levels or number of experiments, to work on another 
% model or with another model space. The script also serves as a good 
% starting point if translation to another script language would be 
% preferred.
%
% The problem is saved to a file (default name is "myProblem")
function GenerateProblem()
close all;
clc;
format compact;
doPlot=1; % 0=no plotting; 1=plot data for all experiments

% Problem name
problemName='myProblem';
delete(problemName);% delete old copy

% Initial conditions for three experimnets (rows)
y0=[3 2 1 0.01 0.01
    4 5 1 0.01 0.01
    1 3 1 0.01 0.01];
[Nexp Nvar]=size(y0);

% Additive noise with standard devation proportional (with constant 0.1) to y
% For each variable
gaussianNoise=[0 0 0 0 0];

% Sampling times
samplingTimes = [0:0.25:3];

% Save to file
diary(problemName);

disp('format version = 1.0');
disp(' ');
disp(['begin problem ' problemName]);
disp(' ');
disp('type = reactionKinetics');
disp(['date = ' datestr(now)]);
disp('url = www.odeidentification.org');
disp(' ');
disp('// REACTION TYPES');
disp(' ');
disp('reaction_1');
disp('has name = uniMolecularMassAction');
disp('has localVariableName_1 = X1');
disp('has localParameterName_1 = k1');
disp('has equation = k1*X1');
disp(' ');
disp('reaction_2');
disp('has name = biMolecularMassAction');
disp('has localVariableName_1 = X1');
disp('has localVariableName_2 = X2');
disp('has localParameterName_1 = k1');
disp('has equation = k1*X1*X2');
disp(' ');

disp(' ');
disp('// VARIABLES');
disp('variable_1 has name = x1 is inputVariable');
disp('variable_2 has name = x2 is inputVariable');
disp('variable_3 has name = x3 is dependent');
disp('variable_4 has name = x4 is dependent');
disp('variable_5 has name = x5 is dependent');

disp(' ');
disp('// RANGES');
disp(' ');
disp('range_1 has lowerBound =  0.00000000E+00 has upperBound =  0.50000000E+02');
disp('range_2 has lowerBound = -0.50000000E+02 has upperBound =  0.00000000E+00');
disp(' ');

disp(' ');
disp('// VARIABLE SETS');
disp(' ');
disp('memberOfSet_1 variable_1 variable_2 variable_3 variable_4 variable_5');
disp('memberOfSet_2 variable_4 variable_5');
disp('memberOfSet_3 variable_3 variable_5');
disp('memberOfSet_4 variable_3 variable_4');
disp(' ');

disp(' ');
disp('// MODEL SPACE');
for ivar=3:Nvar
    disp(' ');
    disp(['// MODEL SPACE OF VARIABLE' int2str(ivar)]);
    disp(' ');
    disp(['possibleReaction_1 of variable_' int2str(ivar)]);
    disp('has type = uniMolecularMassAction');
    disp(['has spaceOfVariable X1 = memberOfSet_' int2str(ivar-1)]);
    disp('has rangeOfParameter k1 = range_1');
    
    disp(' ');
    disp(['possibleReaction_2 of variable_' int2str(ivar)]);
    disp('has type = uniMolecularMassAction');
    disp(['has spaceOfVariable X1 = variable_' int2str(ivar)]);
    disp('has rangeOfParameter k1 = range_2');
    
    disp(' ');
    disp(['possibleReaction_3 of variable_' int2str(ivar)]);
    disp('has type = biMolecularMassAction');
    disp(['has spaceOfVariable X1 = memberOfSet_' int2str(ivar-1)]);
    disp('has spaceOfVariable X2 = memberOfSet_1');
    disp('has rangeOfParameter k1 = range_1');
    
    disp(' ');
    disp(['possibleReaction_4 of variable_' int2str(ivar)]);
    disp('has type = biMolecularMassAction');
    disp(['has spaceOfVariable X1 = variable_' int2str(ivar)]);
    disp('has spaceOfVariable X2 = memberOfSet_1');
    disp('has rangeOfParameter k1 = range_2');
end
disp(' ');


disp(' ');
disp('// ERROR FUNCTION');
disp(' ');
disp('errorFunction');
disp('has type = minusLogLikelihoodPlusLambdaK');
disp('has equation = -L+lambda*K');
disp('has lambda =   1.');
disp(' ');



disp(' ');
for i=1:Nexp
if sum(gaussianNoise)==0.0
disp(['experiment_' int2str(i) ' has name = e' int2str(i) ' has perfectData']);
else
disp(['experiment_' int2str(i) ' has name = e' int2str(i)]);
end
end
disp(' ');


for i=1:Nexp
    disp(' ');
    disp(['// EXPERIMENT ' int2str(i)]);
    
    [t,y] = ode15s(@model,samplingTimes,y0(i,:));
    for j=1:Nvar
        noise=gaussianNoise(j).*randn(length(samplingTimes),1).*y(:,j);
        y(:,j)=y(:,j)+noise; % samples
        stdev(:,j)=y(:,j).*gaussianNoise(j); % "estimate" of standard deviation
% Default non-zero value if noise-free data
        if stdev(:,j)==0
            stdev(:,j)=0.01.*ones(length(samplingTimes),1);
        end
        % Initial bounds for each variable in each experiment
        lb(i,j)=y(1,j)-2*stdev(1,j);
        ub(i,j)=y(1,j)+2*stdev(1,j);
        if (lb<0)
            lb=0;
        end   
    end
    
    if doPlot>0
        figure(i)
        plot(t,y(:,1),'k--');
        hold on;
        plot(t,y(:,2),'b--');
        plot(t,y(:,3),'ro-');
        plot(t,y(:,4),'ms-');
        plot(t,y(:,5),'gd-');
        legend('Y_1 (input)','Y_2 (input)','Y_3','Y_4','Y_5');
        xlabel('Time');
        ylabel('Value');
        title(['Experiment ' int2str(i)]);
    end
    
    for j=1:length(samplingTimes)
        disp(['sample_' int2str(j) ' of experiment_' int2str(i)]);
        disp(['has time =  ' num2str(samplingTimes(j))]);
        disp(['has variable_ =  ' num2str(y(j,:))]);
        disp(['has sdev of variable_ = ' num2str(stdev(j,:))]);
        disp(' ');
    end
end


disp(' ');
disp('// INITIAL BOUNDS');
disp(' ');
for i=1:Nexp
    for j=3:Nvar
        disp(['variable_' int2str(j) ' of experiment_' int2str(i)]);
        disp(['has lowerInitialBound =  '   num2str(lb(i,j))]);
        disp(['has upperInitialBound =  '   num2str(ub(i,j))]);
        disp(' ');
    end
end

disp('// INITIAL SOLUTION (USUALLY ABSENT)');
disp(['end problem ' problemName]);

diary off
end



% The ordinary differential equations
% (all parameters are one)
function dy = model(t,y)
dy = zeros(5,1);
dy(1)=0; % independent
dy(2)=0; % independent
dy(3) = -y(3)+y(1)*y(4);
dy(4) = y(3)-y(1)*y(4)+y(5)-y(2)*y(4);
dy(5) = -y(5)+y(2)*y(4);
end

