/** 
    ODEidon, a program for structural identification of ODEs
    Copyright (C) 2011 Peter Gennemark, Dag Wedelin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Peter Gennemark, Dag Wedelin
             www.odeidentification.org
*/


package Java.generateFortran;

import java.io.*;
import Java.debug.*;
import Java.problem.*;

/**
 * Generates the Fortan program
 * @author      Peter Gennemark, www.odeidentification.org
 * @version     1.0                   
 * @since       1.0
 */
public class GenerateFortran {
    
    String directory;
 
    /**
     * Generates the Fortan program in directory for a certain problem
     * @param dir the directory
     * @param p the problem
     */
    public GenerateFortran(String dir,Problem p){
	this.directory=dir;
	GenerateFortran_spline.generate(directory,p);
	GenerateFortran_readProblem.generate(directory,p);
	GenerateFortran_myReactions.generate(directory,p);
	String fromDirectory= System.getProperty("user.dir") + 
	    System.getProperty("file.separator")+
	    "F_codeSkeleton"+
	    System.getProperty("file.separator");
	generate(p,fromDirectory,directory);
    }
      
    /**
     * Print max sizes of arrays in Fortran program
     * @param out bufferedWriter to write in
     * @param p the problem
     */
    public static void printMaxDefinition(BufferedWriter out,Problem p) throws IOException{

	out.write("      integer maxexp,maxvar,maxpts,maxr,maxv,maxk\n");
	out.write("      parameter (maxexp="+p.getNumberOfExperiments()+                 
		  ",maxvar="+p.getNumberOfVariables()+
		  //",maxpts="+p.get_maxpts()+
		  ",maxpts=201"+
		  ",maxr=10"+//max number of reactions per variable
		  ",maxv="+p.get_maxv()+
		  ",maxk="+p.get_maxk()+")\n");
    }
    

    /**
     * Print model definition in Fortran program
     * @param out bufferedWriter to write in
     * @param include_modkinit true if initial values are parameters, otherwise false
     */
    public static void printModelDefinition(BufferedWriter out, 
					    boolean include_modkinit) throws IOException{
	out.write("C     ------------- BEGIN MODEL DEFINITION -------------------------\n");
	out.write("      integer modtype(maxvar,maxr)\n");
	out.write("C     Type of reaction [1...maxr] for variable [1...maxvar]\n");
	out.write("      \n");
	out.write("      double precision modk(maxvar,maxr,maxk,3) \n");
	out.write("C     Last index=1: Param. [1...maxk] for reaction [1...maxr] in var. [1...maxvar]\n");
	out.write("C     Last index=2: LB for param. [1...maxk] for reaction [1...maxr] in var. [1...maxvar]\n");
	out.write("C     Last index=3: UB for param. [1...maxk] for reaction [1...maxr] in var. [1...maxvar]\n");
	out.write("      \n");
	if (include_modkinit){
	    out.write("      double precision modkinit(maxexp,maxvar,3) \n");
	    out.write("C     Last index=1: Init conc for experiment [1...maxexp], variable [1...maxvar]\n");
	    out.write("C     Last index=2: Lower bound for experiment [1...maxexp], variable [1...maxvar]\n");
	    out.write("C     Last index=3: Upper bound for experiment [1...maxexp], variable [1...maxvar]\n");
	    out.write("      \n");
	}
	out.write("      integer modn(maxvar)\n");
	out.write("C     Number of reactions for variable [1...maxvar]\n");
	out.write("      \n");
	out.write("      integer modv(maxvar,maxr,maxv)\n");
	out.write("C     Variable [1...maxv] for reaction [1...maxr] in variable [1...maxvar]\n");
	out.write("      \n");
	out.write("      logical dependent(maxvar)\n");
	out.write("C     Which variables are dependent\n");
	out.write("C     ------------- END MODEL DEFINITION -------------------------\n");
    }


    /**
     * Print definition for temporary model variables in Fortran program
     * @param out bufferedWriter to write in
     * @param include_modkinit true if initial values are parameters, otherwise false
     */
    public static void printTempModelDefinition(BufferedWriter out,
						boolean include_modkinit) throws IOException{
	out.write("C     ------------- BEGIN TEMP MODEL DEFINITION -------------------------\n");
	out.write("      integer tmodtype(maxvar,maxr)\n");
	out.write("      double precision tmodk(maxvar,maxr,maxk,3) \n");
	if (include_modkinit){
	    out.write("      double precision tmodkinit(maxexp,maxvar,3) \n");
	}
	out.write("      integer tmodn(maxvar)\n");
	out.write("      integer tmodv(maxvar,maxr,maxv)\n");
	out.write("C     ------------- END TEMP MODEL DEFINITION -------------------------\n");
    }


    /**
     * Print definition for second temporary model variables in Fortran program
     * @param out bufferedWriter to write in
     * @param include_modkinit true if initial values are parameters, otherwise false
     */
    public static void printTempModelDefinition2(BufferedWriter out,
						 boolean include_modkinit) throws IOException{
	out.write("C     ------------- BEGIN TEMP MODEL DEFINITION 2-------------------------\n");
	out.write("      integer t2modtype(maxvar,maxr)\n");
	out.write("      double precision t2modk(maxvar,maxr,maxk,3) \n");
	if (include_modkinit){
	    out.write("      double precision t2modkinit(maxexp,maxvar,3) \n");
	}
	out.write("      integer t2modn(maxvar)\n");
	out.write("      integer t2modv(maxvar,maxr,maxv)\n");
	out.write("C     ------------- END TEMP MODEL DEFINITION 2-------------------------\n");
    }


    /**
     * Print definition for the experiments in Fortran program
     * @param out bufferedWriter to write in
     * @param include_edata2 true if the variable edata2 is included, otherwise false
     * @param include_spline true if the variable spline is included, otherwise false
     * @param include_useip true if the variable useip is included, otherwise false
     */
    public static void printExperimentDefinition(BufferedWriter out,
						 boolean include_edata2,
						 boolean include_spline,
						 boolean include_useip) throws IOException{
	out.write("C     ------------- BEGIN EXPERIMENT DEFINITION -----------------------\n");
	out.write("      double precision edata(maxexp,maxvar+1,2*maxpts) \n");
	if (include_edata2) 
	    out.write("      double precision edata2(maxexp,maxvar+1,2*maxpts)\n");   
	if (include_spline) 
	    out.write("      integer spline(maxexp,maxvar) ! 0=linear interpolation, 1=cubic smoothing spline \n");
	if (include_edata2 && include_spline)
	    out.write("      integer spline2(maxexp,maxvar) ! 0=linear interpolation, 1=cubic smoothing spline \n");
	if (include_useip) 
	    out.write("      logical useip(maxvar)     ! use interpolated data or simulated data in simulateSingle\n");
	out.write("C     ------------- END EXPERIMENT DEFINITION -------------------------\n");
    }

    /**
     * Pretty printout in Fortran program, no line number in column 1-6
     * @param out bufferedWriter to write in
     * @param s the String to write
     * @param indent number of indented columns
     */
    public static void prettyPrint(BufferedWriter out, 
				   String s, 
				   int indent) throws IOException{
	prettyPrint(out,s,indent,0);
    }


   /**
     * Pretty printout in Fortran program
     * @param out bufferedWriter to write in
     * @param s the String to write
     * @param indent number of indented columns
     * @param linenumber line number in column 1-6
     */
    public static void prettyPrint(BufferedWriter out, 
				   String s, 
				   int indent,
				   int linenumber) throws IOException{
	int INDENTSTEP=3;
	int MAX=70;
	if (linenumber==0) out.write("      ");
	else if (linenumber<10) out.write(" "+linenumber+"    ");
	else if (linenumber<100) out.write(" "+linenumber+"   ");
	else if (linenumber<1000) out.write(" "+linenumber+"  ");
	else if (linenumber<10000) out.write(" "+linenumber+" ");
	for (int j=0;j<indent*INDENTSTEP;j++) out.write(" ");
	int i=0;
	for (int j=0;;j++) {
	    if (j>0){ 
		out.write("     *");
		for (int jj=0;jj<indent*INDENTSTEP;jj++) out.write(" ");	
		out.write("  ");
	    }
	    if (s.substring(i,s.length()).length()>(MAX-6-indent*INDENTSTEP)){
		out.write(s.substring(i,i+(MAX-6-indent*INDENTSTEP))+"\n");
		i+=MAX-6-indent*INDENTSTEP;
	    }
	    else{
		out.write(s.substring(i,s.length())+"\n");
		break;
	    }
	}
    }
    
   /**
     * Java double to Fortran double, replace E by d, or add d0 at end 
     * @param d the input real number of Java type double
     * @return a String representing the modfied number
     */
    public static String fixExponent(double d){
	String s=  Double.toString(d); 
	if (s.contains("E")) 
	    s=s.replace("E","d");
	else s+="d0";
	return s;
    }


    /**
     * Adapt an equation to the Fortran parameter estimation routine
     * @param ein a String representing the equation of a certain reaction type
     * @param r the current Reaction type
     * @return a String representing the modfied equation
     */
    public static String adaptEquationToParaest(String ein,
						ReactionType r){
	String e=ein.replace("^","**");
	e=e.replace("ln","log");
	for (int j=0;j<r.getNumberOfLocalParameters();j++){
	    e=e.replace("k"+(j+1),"X(ipara+"+j+")");
	}
	for (int j=0;j<r.getNumberOfLocalVariables();j++){
	    e=e.replace("X"+(j+1),"edata(iexp,modv(ivar,ireact,"+(j+1)+")+1,k)");
	}
	return e;
    }

    /**
     * Adapt an equation to the Fortran simulation routine
     * @param ein a String representing the equation of a certain reaction type
     * @param r the current Reaction type
     * @return a String representing the modfied equation
     */
    public static String adaptEquationToSim(String ein,ReactionType r){
	String e=ein.replace("^","**");
	e=e.replace("ln","log");
	for (int j=0;j<r.getNumberOfLocalParameters();j++){
	    e=e.replace("k"+(j+1),"modk(i,ireact,"+(j+1)+",1)");
	}
	for (int j=0;j<r.getNumberOfLocalVariables();j++){
	    e=e.replace("X"+(j+1),"y(modv(i,ireact,"+(j+1)+"))");
	}
	return e;
    }


    /**
     * Adapt an equation to the Fortran routine for simulation of a single variable
     * @param ein a String representing the equation of a certain reaction type
     * @param r the current Reaction type
     * @return a String representing the modfied equation
     */
    public static String adaptEquationToSimSingle(String ein,ReactionType r,int ivar){
	String e=ein.replace("^","**");
	e=e.replace("ln","log");
	for (int j=0;j<r.getNumberOfLocalParameters();j++){
	    e=e.replace("k"+(j+1),"modk(i,ireact,"+(j+1)+",1)");
	}
	for (int j=0;j<r.getNumberOfLocalVariables();j++){
	    if (ivar==j) e=e.replace("X"+(j+1),"y(1)");
	    else e=e.replace("X"+(j+1),"ytemp("+(j+1)+")");
	}
	return e;
    }


       
    /**
     * Generate Fortran files that are based on code skeletons in files 
     * @param p the problem
     * @param fromDirectory directory of code skeletons
     * @param toDirectory directory of generated code
     */
    public void generate(Problem p,String fromDirectory,String toDirectory) {

	
	String[] files={"main.f","alg.f","alg_paraest.f",
			"err.f","est.f","estvar.f",
			"estvars.f","exp.f",
			"funjac1.f","funjac1general.f",
			"funjac4.f","funjac4general.f",
			"model.f","sim.f"};


	for (int i=0;i<files.length;i++) {
	    Debug.print(this,"generate",files[i]);
	    
	    File filefrom = new File(fromDirectory + files[i]);
	    File fileto = new File(toDirectory + files[i]);
	    int line_no=0;
	    
	    try {
		BufferedReader in = new BufferedReader(new FileReader(filefrom));
		BufferedWriter out = new BufferedWriter(new FileWriter(fileto));
		out.write("C     This file was automatically generated by GenerateFortran.generate()\n");
		out.write("C     on "+(new java.util.Date())+"\n");
		out.write("C\n");
		for (;;) { 
		   if (!in.ready()) break;
		   String s=in.readLine();
		   
		   // Replace sizes of data-structures
		   s=s.replace("maxexp=20","maxexp="+p.getNumberOfExperiments());
		   s=s.replace("maxvar=19","maxvar="+p.getNumberOfVariables());
		   s=s.replace("maxv=3","maxv="+p.get_maxv());
		   s=s.replace("maxk=5","maxk="+p.get_maxk());
		   s=s.replace("maxtypes=50","maxtypes="+p.getNumberOfReactionTypes());
		   
		   // Setting for dn2gb
		   // iLIV = 82 + 4*P
		   // iLV = 105 + P*(N + 2*P + 21) + 2*N
		   if (files[i].equals("estvar.f")) {
		       int P=p.getMAX_PARAMETERS_PER_ODE();
		       int N=p.getMaxNumberOfSamplesPerVariable();
		       s=s.replace("35",""+P); // max number of para's
		       s=s.replace("parameter(LIV=222,LV=10500,LR=920)","parameter(LIV="+(82+4*P)+
				   ",LV="+(105+P*(N+2*P+21)+2*N)+",LR="+N+") ! automatically set");
		   }
		   else if (files[i].equals("est.f")) {
		       int P=p.getMAX_PARAMETERS_PER_ODE()*p.getNumberOfDependentVariables();
		       int N=p.getMaxNumberOfSamplesPerVariable()*p.getNumberOfDependentVariables(); 
		       s=s.replace("400",""+P); // max number of para's
		       s=s.replace("parameter(LIV=1200,LV=600000,LR=12000)","parameter(LIV="+(82+4*P)+
				   ",LV="+(105+P*(N+2*P+21)+2*N)+",LR="+N+") ! automatically set 1");
		       s=s.replace("parameter(LIV=3000,LV=1500000,LR=30000)","parameter(LIV="+(82+4*P)+
				   ",LV="+(105+P*(N+2*P+21)+2*N)+",LR="+N+") ! automatically set 2");
		   }
		   else if (files[i].equals("estvars.f")) {
		       int P=p.getMAX_PARAMETERS_PER_ODE()*p.getNumberOfDependentVariables();
		       int N=p.getMaxNumberOfSamplesPerVariable()*p.getNumberOfDependentVariables();
		       s=s.replace("400",""+P); // max number of para's
		       s=s.replace("parameter(LIV=1200,LV=800000,LR=6000)","parameter(LIV="+(82+4*P)+
				   ",LV="+(105+P*(N+2*P+21)+2*N)+",LR="+N+") ! automatically set");
		   }
		   else if (files[i].equals("funjac1.f")) {
		       int P=p.getMAX_PARAMETERS_PER_ODE()*p.getNumberOfDependentVariables();
		       s=s.replace("400",""+P); // max number of para's
		   }
		   else if (files[i].equals("funjac1general.f")) {
		       int P=p.getMAX_PARAMETERS_PER_ODE()*p.getNumberOfDependentVariables();
		       s=s.replace("400",""+P); // max number of para's
		   }
		   else if (files[i].equals("funjac4.f")) {
		       int P=p.getMAX_PARAMETERS_PER_ODE()*p.getNumberOfDependentVariables();
		       s=s.replace("400",""+P); // max number of para's
		   }
		   else if (files[i].equals("funjac4general.f")) {
		       int P=p.getMAX_PARAMETERS_PER_ODE()*p.getNumberOfDependentVariables();
		       s=s.replace("400",""+P); // max number of para's
		   }
                   else if (files[i].equals("alg.f")) {
                       if (p.isSSystem()) s=s.replace("estgeneral","est"); 
                   }
                   else if (files[i].equals("alg_paraest.f")) {
                       if (p.isSSystem()) s=s.replace("estgeneral","est"); 
                   }
                   else if (files[i].equals("err.f")) {
		       String eq=p.getErrorFunction().getEquation();
                       s=s.replace("-L+lambda*K",eq); 
                   }

		   out.write(s+"\n");
		   line_no++;
	       }  
	       in.close();
	       out.close();
	    } catch(IOException ioe) {    
		String errorMsg="Error reading/writing file " + files[i] + ", line number: " + line_no;   
		Debug.error(this,"generate",errorMsg); 
		System.exit(0);	
	    } 
	}
    }
    
    
    // public static void main(String[] argv){ 
    // 	Debug.ON = false;
    // 	Debug.ERROR = false;
    // 	Debug.WARNING = false;
 
    // 	// File system variables
    // 	String filetree = System.getProperty("user.dir") + 
    // 	    System.getProperty("file.separator");
	
    // 	// Problem file
    // 	File file=new File(filetree+argv[0]);
	
    // 	Problem p=new Problem();
    // 	p.read(file,0);    
	
    // 	GenerateFortran gf = new  GenerateFortran("F_generatedCode"+
    // 						  System.getProperty("file.separator"),
    // 						  p);
    // 	Debug.stat("GenerateFortran","main","Code for problem "+argv[0]+
    // 		   " is generated in directory F_generatedCode.");
    // }
    
    
}
