/** 
    ODEidon, a program for structural identification of ODEs
    Copyright (C) 2011 Peter Gennemark, Dag Wedelin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Peter Gennemark, Dag Wedelin
             www.odeidentification.org
*/

package Java.gui;

import Java.debug.*;
import java.lang.Math.*;


/** Smoothing spline interpolation 
  *
  * This class provides an interpolation/approximation facility to work with data.<p>
  * Partly this code is a modification of fortran <i>subroutine smooth</i> (file smooth.f)
  * from <i>pppack</i> library at <a href="http://www.netlib.org">Netlib</a>
  * (editor: Eric Grosse, master: 
  * <a href="http://netlib.bell-labs.com">netlib.bell-labs.com</a>)
  * Original code written by Alexander Sokolov,
  * Department of Systems Ecology, Stockholm University, 1997
  * Modified by Peter Gennemark, Chalmers, May 2003
  * @author      Peter Gennemark, www.odeidentification.org
  * @version     1.0                   
  * @since       1.0
  */
public class Pppack_smooth {
  
  /** Number of data points */
  public int lastElement;
  /** Array of data abscissae, assumed to be strictly increasing */
  public double[] x;
  /** Array of data ordinates, corresponded to abscissae */
  public double[] y;
  /** Array of smoothed ordinates, corresponded to abscissae */
  public double[] f;
  
  public double[] d1f, d2f, d3f;  // first three derivatives at the left end of the interval
  double[] dy;             // estimate of uncertainty in data (positive)
  double[][] v;

  
  int klo=0;// Remember previous call to the method get,
  int khi=0;// to speed up bisection.



  /** Constructs the spline class.
    * @param x x-values
    * @param y y-values
    * @param dy - estimate of uncertainty in data (positive)
    */
  public Pppack_smooth(double[] x,
		       double[] y,
		       double[] dy) { 
    this.x=x;
    this.y=y;
    this.dy=dy;
    f=new double[x.length];
    d1f=new double[x.length];
    d2f=new double[x.length];
    d3f=new double[x.length];
    v=new double[x.length][7];
    lastElement=x.length-1;
  }
  

  
  private void SetupQ()
    { v[0][3]=x[1]-x[0];
    for (int i=1; i < lastElement; i++)
      { v[i][3]=x[i+1]-x[i];
      v[i][0]=dy[i-1]/v[i-1][3];
      v[i][1]=-dy[i]/v[i][3]-dy[i]/v[i-1][3];
      v[i][2]=dy[i+1]/v[i][3];
      }
    v[lastElement][0]=0.0;
    for (int i=1; i < lastElement; i++)
      v[i][4]=Math.pow(v[i][0],2.0)+Math.pow(v[i][1],2.0)+Math.pow(v[i][2],2.0);
    for (int i=2; i < lastElement; i++)
      v[i-1][5]=v[i-1][1]*v[i][0]+v[i-1][2]*v[i][1];
    v[lastElement-1][5]=0.0;
    for (int i=3; i < lastElement; i++)
      v[i-2][6]=v[i-2][2]*v[i][0];
    v[lastElement-2][6]=0.0;
    v[lastElement-1][6]=0.0;
    // construct  q-transp. * y  in  d3f.
    double prev=(y[1]-y[0])/v[0][3];
    for (int i=1; i < lastElement; i++)
      { double diff=(y[i+1]-y[i])/v[i][3];
      d3f[i]=diff-prev;
      prev=diff;
      }
    }
  
  private void Chol1D(double p)
    { double temp=6.0*(1.0-p);
    for (int i=1; i < lastElement; i++)
      { v[i][0]=temp*v[i][4]+2.0*p*(v[i-1][3]+v[i][3]);
      v[i][1]=temp*v[i][5]+p*v[i][3];
      v[i][2]=temp*v[i][6];
      }
    if (lastElement >= 3)
      { // factorization
	for (int i=1; i < lastElement-1; i++)
	  { double ratio=v[i][1]/v[i][0];
	  v[i+1][0]-=ratio*v[i][1];
	  v[i+1][1]-=ratio*v[i][2];
	  v[i][1]=ratio;
	  ratio=v[i][2]/v[i][0];
	  v[i+2][0]-=ratio*v[i][2];
	  v[i][2]=ratio;
	  }
	// forward substitution
	d2f[0]=0.0;
	d2f[1]=d3f[1];
	v[0][2]=0.0;
	for (int i=1; i < lastElement-1; i++)
	  d2f[i+1]=d3f[i+1]-v[i][1]*d2f[i]-v[i-1][2]*d2f[i-1];
	// back substitution
	d2f[lastElement]=0.0;
	d2f[lastElement-1]=d2f[lastElement-1]/v[lastElement-1][0];
	for (int i=lastElement-2; i > 0; i--)
        d2f[i]=d2f[i]/v[i][0]-d2f[i+1]*v[i][1]-d2f[i+2]*v[i][2];
      }
    else
      { d2f[0]=0.0;
      d2f[1]=d3f[1]/v[1][0];
      d2f[2]=0.0;
      }
    // construct q*u
    double prev=0.0;
    for (int i=1; i <= lastElement; i++)
      { f[i]=(d2f[i]-d2f[i-1])/v[i-1][3];
      f[i-1]=f[i]-prev;
      prev=f[i];
      }
    f[lastElement]=-f[lastElement];
    }
  
  


  /** Constructs the cubic smoothing spline to given data (x(),y()).
    * This approximation has a second derivative as small as possible
    * while sum(((y(i)-f(x(i)))/dy(i))^2) < maxMeanDistance
    * @param maxMeanDistance is upper bound on the discrete weighted mean square distance of
    * the approximation from the data
    */
  public void calcSpline(double maxMeanDistance) { 
    if (lastElement > 0) { 
      if (lastElement < 2 | maxMeanDistance < -0.9) { 
	for (int i=0; i < lastElement; i++) { 
	  f[i]=y[i];
	  d1f[i]=(y[i+1]-y[i])/(x[i+1]-x[i]);
	  d2f[i]=0.0;
          d3f[i]=0.0;
	}
        f[lastElement]=y[lastElement];
      }
      else { 
	int  maxIter=100;   // Usually it takes 3 to 5 iterations
        double sfq, p=0.0, q=0.0;
        SetupQ();
        if (maxMeanDistance > 0.0){ 
	  p=0.0;
          Chol1D(p);
          sfq=0.0;
          for (int i=0; i <= lastElement; i++)
            sfq+=Math.pow(f[i]*dy[i], 2.0);
          sfq=sfq*36.0;
          if (sfq > maxMeanDistance){ 
	    double utru=0.0;
            for (int i=1; i <= lastElement; i++)
              utru+=v[i-1][3]*(d2f[i-1]*(d2f[i-1]+d2f[i])+Math.pow(d2f[i], 2.0));
            double ooss=1.0/Math.sqrt(maxMeanDistance);
            double oosf=1.0/Math.sqrt(sfq);
            q=-(oosf-ooss)*sfq/(6.0*utru*oosf);
            double prevq=0.0;
            double prevsf=oosf;
            boolean OK;
            int iter=0;
            do { 
	      p=q/(1.0+q);
              Chol1D(p);
              iter+=1;
              sfq=0.0;
              for (int i=0; i <= lastElement; i++)
                sfq+=Math.pow(f[i]*dy[i], 2.0);
              sfq=(sfq*36.0)/Math.pow(1.0+q, 2.0);
              OK=Math.abs(sfq-maxMeanDistance) <= 0.01*maxMeanDistance;
              if (!OK){ 
		oosf=1.0/Math.sqrt(sfq);
                double change=(q-prevq)/(oosf-prevsf)*(oosf-ooss);
                prevq=q;
                q-=change;
                prevsf=oosf;
		}
	    }
            while (!OK && iter <= maxIter);
            p=q/(1.0+q);
	  }
	}
        else { 
	  p=1.0;
          Chol1D(p);
          sfq=0.0;
          q=0.0;
	}
        double temp=6.0/(1.0+q);
        temp= q > 0.0 ? temp : 0.0;
        for (int i=0; i <= lastElement; i++)
          f[i]=y[i]-temp*Math.pow(dy[i], 2.0)*f[i];
        for (int i=0; i <= lastElement; i++)
          d2f[i]*=6.0*p;
        for (int i=0; i < lastElement; i++){ 
	    d3f[i]=(d2f[i+1]-d2f[i])/v[i][3];
	    d1f[i]=(f[i+1]-f[i])/v[i][3]-(d2f[i]+d3f[i]/3.0*v[i][3])/2.0*v[i][3];
	}
	//Debug.print(this,"calcSpline","p="+p);
      }
    }
  }
  
  /**
    * Constructs an interpolation/approximation function to given data (x(),y()).
    * @param isApproximation is type of a function. If it is true
    * the approximative function will be made, otherwise -- a linear function.
    * @param maxMeanDistance is upper bound on the discrete weighted mean square distance of
    * the approximation from the data
    */
  public void calcSpline(boolean isApproximation,
                         double maxMeanDistance){
    if (isApproximation)
      calcSpline(maxMeanDistance);
    else
      calcSpline(-1.0);  // Linear interpolation
  }
  
  
  /** Calculates value of the function.
    * @param coord is a value of abscissa for searching function's value.
    * @return value of the interpolation/approximation function for coord value of abscissa
    * or 0 for unappropriate abscissa's value.
    */
  public double get(double coord){ 
    Debug.print(this,"get(double coord)","Use get(double coord, "+
		"int jderiv) instead, it is more optimized.");
    double temp;
    //double result=emptyCode-0.5;
    double result=0;
    boolean find=false;
    if (lastElement > 0){ 
      int i;
      for (i=0; (i < lastElement) &&
	     (!(find=(coord >= x[i]) && (coord <= x[i+1]))); i++){};
      if (find){ 
	temp=coord-x[i];//delta x
	result=f[i]+temp*(d1f[i]+temp*(d2f[i]+temp*d3f[i]/3.0)/2.0);
      }
    }
    return result;
  }


  
  /** Calculates value of the function.
    * @param coord is a value of abscissa for searching function's value.
    * @param jderiv 0 or 1, function value (0) or first derivative (1)
    * @return value of the interpolation/approximation function for coord value of abscissa
    * or 0 for unappropriate abscissa's value.
    */
  public double get(double coord, 
		    int jderiv){ 
    double h;
    double result=0;
    //boolean find=false;
    if (lastElement > 0){ 
      
      
      // We will find the right place in the table by means of
      // bisection. This is optimal if sequential calls to this
      // routine are at random values of x. If sequential calls
      // are in order, and closely spaced, one would do better
      // to store previous values of klo and khi and test if
      // they remain appropriate on the next call.

      // Initial test, are we lucky and have sequential calls or one step forward
      // in the case of one call for value and one for derivative
      if (khi<lastElement) khi++; // one step forward
      if (x[klo] <= coord && x[khi] > coord ) {
	// success, bisection on a small interval
	//System.out.println("Shortcut success, klo="+klo+" khi="+khi);
      }
      else{ // normal bisection
	klo=0;
	khi=lastElement;
	//System.out.println("Normal bisection");
      }
      int k;
      while (khi-klo > 1) {
	k=(khi+klo) >> 1;
	if (x[k] > coord) khi=k;
	else klo=k;
      } //klo and khi now bracket the input time of x.
      int i=klo;
      

      h=coord-x[i];
      if (jderiv==0)
	result=f[i]+h*(d1f[i]+h*(d2f[i]+h*d3f[i]/3.0)/2.0);
      else if (jderiv==1){
	result=d1f[i]+h*(d2f[i]+h*(d3f[i])/2.0);
	}
      else
	Debug.error(this,"get","jderiv must be 0 or 1, jderiv="+jderiv);
    }
    return result;
  }

    /**
     * Get value of variable f at position val for a certain derivative [0,1,2,3}.
     * @param index the index
     * @param val the position
     * @return the value
     */
    public double getf(int index,int val){
	if (val>=0 && val<x.length){
	    if (index==0) 
		return f[val];
	    else if (index==1) 
		return d1f[val];
	    else if (index==2) 
		return d2f[val];
	    else if (index==3) 
		return d3f[val];
	}
	else{
	    Debug.error(this,"getf","val must be 0<=val<x.length, val="+val+" x.length="+x.length);

	}
	Debug.error(this,"getf","index must be 0, 1, 2 or 3, index="+index);
	return -1;
    }
  
  
  // public static void main(String[] argv){


  //     int k1=10;
  //     int k2=20;
  //     int p= (k1+k2) >> 1;
  //     System.out.println(p);

  //     for (k1=11;k1<22;k1++){
  // 	  p= (k1+k2) >> 1;
  // 	  System.out.println(k1+" "+p);
  //     }

 
  // }
 
}

   
