/** 
    ODEidon, a program for structural identification of ODEs
    Copyright (C) 2011 Peter Gennemark, Dag Wedelin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Peter Gennemark, Dag Wedelin
             www.odeidentification.org

*/

package Java.problem;
import java.util.*;
import Java.debug.*;
import java.io.*;
import Java.misc.*;



/**
  * Description of the error function
  * @author Peter Gennemark
  * @version     1.0                   
  * @since       1.0
  */
public class ErrorFunction  {
 
    private String type=null;
    private String equation=null;
    private double lambda=0.0;
     
    public ErrorFunction(){
	
    }


    /** Read error function from input file in form of a buffered reader
     * @param commands vector of strings on first line to read
     * @param in buffered reader
     * @param line_no current line in the buffered reader (i.e., problem file)
     * @param debug debug level
     */
   public int read(String[] commands,
		   BufferedReader in,
		   int line_no,
		   int debug){
       boolean readlambda=false; 
	try {
	    for (;;) {
		String s= in.readLine();
		line_no++;
		commands = Misc.split(s,line_no,debug);
		if (commands==null){
		    if (debug>0) Debug.print(this,"read","empty line");
		    continue;
		}
		else if (commands[0].equals("has") && commands[1].equals("type") &&
			 commands[2].equals("=")) {
		    type=commands[3];
		    if (type.equals("minusLogLikelihoodPlusLambdaK")){
		    }
		    else if (type.equals("minusLogLikelihood")){
			// parameter estimation
			
		    }
		    else{
			Debug.warning(this,"read","Unrecognized type: "+type+", on line="+line_no);
			//System.exit(0); 
		    }
		}
		else if (commands[0].equals("has") && 
			 commands[1].startsWith("equation") && 
			 commands[2].equals("=")){
		    equation=commands[3];
		    if (equation.equals("-L+lambda*K")){
			// standard case
		    }
		    else if (equation.equals("-L")){
			// parameter estimation
			readlambda=true; // lambda not required
		    }
		    else{
			Debug.warning(this,"read","Equation "+equation+" on line="+line_no+" may cause problems in the Fortran program.\nThe error function should be a function of the variables L (log likelihood), lambda (parameter), K  (number of model parameters) and NS (number of samples), and written in Fortran syntax, e.g., -L+lambda*k*log(NS)");
		    }
		}
		else if (commands[0].equals("has") && 
			 commands[1].startsWith("lambda") && 
			 commands[2].equals("=")){
		    lambda=Misc.getDouble(commands[3],line_no);
		    if (lambda<0.0){
			Debug.error(this,"read","Lambda should be >= 0: "+
				    lambda+", on line="+line_no);
			System.exit(0); 
		    }
		    readlambda=true;
		}
		else {
		    Debug.error(this,"read","Unexpected command: "+commands[0]+", on line="+line_no);
		    System.exit(0); 
		}
		
		if (type!=null && equation!=null && readlambda) break;
	    }
	    
	}
	catch (IOException ioe) {
	    Debug.error(this,"read","IOExpection on line="+line_no);
	    System.exit(0); 
	}
	return line_no;
    }

    /** Get the type
     * @return the type
     */ 
    public String getType(){
	return type;
    }
    
    /** Get the lambda parameter
     * @return lambda
     */ 
    public double getLambda(){
	return lambda;
    }
    
    /** Get the equation in string form
     * @return the equation
     */ 
    public String getEquation(){
	return equation;
    }

    /** Get string representation
     * @return the string
     */ 
    public String toString(){
	String s="Error function "+type+" equation: "+equation+", lambda="+lambda+"\n";
	return s;
    }


    //public static void main(String[] argv) {

    //}
  
}
