/** 
    ODEidon, a program for structural identification of ODEs
    Copyright (C) 2011 Peter Gennemark, Dag Wedelin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Peter Gennemark, Dag Wedelin
             www.odeidentification.org

*/

package Java.problem;
import java.util.*;
import Java.debug.*;
import java.io.*;
import Java.misc.*;


/**
  * Description of an experiment
  * @author Peter Gennemark
  * @version     1.0                   
  * @since       1.0
  */

public class Experiment  {
    private static int totalid=0;
    private int id;

    private Vector<Sample> samples=new Vector<Sample>();
    private String name;
    private boolean perfectData=false;
    private int N_VARIABLES;
    
    private double[] lowerInitialBounds;
    private double[] upperInitialBounds;
  
        
    /** Constricts an experiment
     * @param N_VARIABLES number of variables
     * @param line_no current line in the buffered reader (i.e., problem file)
     */
    public Experiment(int N_VARIABLES,int line_no){
	this.id=totalid++;
	this.N_VARIABLES=N_VARIABLES;
	if (N_VARIABLES<=0) {
	    Debug.constructor(this,"Illegal number of variables"+N_VARIABLES+
			      ", on line="+line_no);
	    System.exit(0); 
	}
	lowerInitialBounds=new double[N_VARIABLES];
	upperInitialBounds=new double[N_VARIABLES];
    }
    
    
    /** Read range from input file in form of a buffered reader
     * @param commands vector of strings on first line to read
     * @param in buffered reader
     * @param line_no current line in the buffered reader (i.e., problem file)
     * @param debug debug level
     */
    public int read(String[] commands,
		    BufferedReader in,
		    int line_no,
		    int debug){
	
	id = Misc.getInteger(commands[0],line_no);
	if (id<=0) {
	    Debug.error(this,"read","id="+id+" should be positive on line="+line_no);
	    System.exit(0);
	}
	if (id!=totalid) {
	    Debug.error(this,"read","range_"+id+" should be range_"+totalid+" on line="+line_no);
	    System.exit(0);
	}
	if (commands.length>4 && commands[1].equals("has") && commands[2].equals("name") &&
	    commands[3].equals("=")) {
	    name=commands[4];
	}
	else {
	    Debug.error(this,"read","Unexpected command: "+commands[1]+" "+
			commands[2]+" "+commands[3]+", on line="+line_no);
	    System.exit(0); 
	}
	if (commands.length>6 && commands[5].equals("has")  && commands[6].equals("perfectData")) {
	    perfectData=true;
	}
   	return line_no;
	
    }



    public String toString() {
	String p= name+" perfect data="+perfectData;
	return p;
    }
      

    /** Get time-points
     * @return vector of time-points
     */
    public double[] getTimeArray(){
	double[] timeArray=new double[getNumberOfSamples()];
	for (int i=0;i<getNumberOfSamples();i++){
	    timeArray[i]=getSample(i).getTime();
	}
	return timeArray;
    }


    /** Get data values for variable i
     * @param ivar the variable i
     * @return vector of data
     */
    public double[] getValueArray(int ivar){
	double[] valueArray=new double[getNumberOfSamples()];
	for (int i=0;i<getNumberOfSamples();i++){
	    valueArray[i]=getSample(i).getValue(ivar);
	}
	return valueArray;
    }

    /** Get data standard deviations for variable i
     * @param ivar the variable i
     * @return vector of standard deviations
     */
    public double[] getSdevArray(int ivar){
	double[] sdevArray=new double[getNumberOfSamples()];
	for (int i=0;i<getNumberOfSamples();i++){
	    sdevArray[i]=getSample(i).getSdev(ivar);
	}
	return sdevArray;
    }
  
  
   /** Get the max data value over all samples
     * @return the max value
     */
    public double getMaxValue(){
	double max=0;
	for (int i=0;i<N_VARIABLES;i++){
	    if (max<getMaxValue(i)) 
		max=getMaxValue(i);
	}
	return max;
    }

   /** Get the min data value over all samples
     * @return the min value
     */
    public double getMinValue(){
	double min=100000000000.0;
	for (int i=0;i<N_VARIABLES;i++){
	    if (min>getMinValue(i)) 
		min=getMinValue(i);
	}
	return min;
    }


   /** Get the max data value over all samples for variable i
     * @param ivar the variable i
     * @return the max value
     */
    public double getMaxValue(int ivar){
	double max=0;
	for (int i=0;i<getNumberOfSamples();i++){
	    if (max<getSample(i).getValue(ivar)) 
		max=getSample(i).getValue(ivar);
	}
	return max;
    }

    /** Get the min data value over all samples for variable i
     * @param ivar the variable i
     * @return the min value
     */
    public double getMinValue(int ivar){
	double min=1000000000.0;
	for (int i=0;i<getNumberOfSamples();i++){
	    if (min>getSample(i).getValue(ivar)) 
		min=getSample(i).getValue(ivar);
	}
	return min;
    }


   /** Get the max time point over all samples
     * @return the max time point
     */
    public double getMaxTime(){
	double max=0;
	for (int i=0;i<getNumberOfSamples();i++){
	    if (max<getSample(i).getTime()) 
		max=getSample(i).getTime();
	}
	return max;
    }

   /** Get the min time point over all samples
     * @return the min time point
     */
    public double getMinTime(){
	double min=1000000000;
	for (int i=0;i<getNumberOfSamples();i++){
	    if (min>getSample(i).getTime()) 
		min=getSample(i).getTime();
	}
	return min;
    }

        
 
    /** Add one sample
     * @param s the sample
     */
    public void addSample(Sample s){
	samples.add(s);
    }
    
    /** Get sample i
     * @param i index of the sample
     * @return the sample
     */
    public Sample getSample(int i){
	return samples.get(i);
    }

    /** Get the number of samples
     * @return the number of samples
     */
    public int getNumberOfSamples(){
	return samples.size();
    }

    /** Indication if data is perfect
     * @return true if experiment has perfect data (artificial test data without noise), o.w. false
     */

    public boolean hasPerfectData(){
	return perfectData;
    }


    /** Add initial bound
     * @param ib the initial bound
     * @param v variable to add initial bound to
     */
    public void addInitialBound(InitialBound ib,Variable v){
	lowerInitialBounds[v.getId()]=ib.getLowerBound();
	upperInitialBounds[v.getId()]=ib.getUpperBound();
    }

    /** Get lower initial bound for variable i
     * @param ivar the variable index
     * @return the lower initial bound
     */
    public double getLowerInitialBound(int ivar){
	return lowerInitialBounds[ivar];
    }

    /** Get upper initial bound for variable i
     * @param ivar the variable index
     * @return the upper initial bound
     */
    public double getUpperInitialBound(int ivar){
	return upperInitialBounds[ivar];
    }


  
  
    // public static void main(String[] argv) {
   
    // 	// File system variables
    // 	String filetree = System.getProperty("user.dir") + 
    // 	    System.getProperty("file.separator");
	
    // 	// Problem file
    // 	File file=new File(filetree+argv[0]);
	
	
    
    // 	Problem p=new Problem();
    // 	p.read(file,1);  
	
    // 	Experiment e =p.getExperiment(0);
    // 	double[] v=e.getValueArray(2);
    // 	Java.gui.Pppack_smooth pp=new Java.gui.Pppack_smooth(e.getTimeArray(),
    // 							     v,e.getSdevArray(2));
    // 	pp.calcSpline(p.getSmoothingParameter());
    // 	for (int i=0;i<13;i++){
    // 	    double t=0.25*i;
    // 	    System.out.println(t+"\t"+pp.get(t,0)+"\t"+v[i]);
    // 	    }

    // }
  
  
}
