/** 
    ODEidon, a program for structural identification of ODEs
    Copyright (C) 2011 Peter Gennemark, Dag Wedelin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Peter Gennemark, Dag Wedelin
             www.odeidentification.org

*/



package Java.problem;

import java.util.*;
import Java.debug.*;
import java.io.*;
import Java.misc.*;

/**
  * Description of a possible reaction
  * @author Peter Gennemark
  * @version     1.0                   
  * @since       1.0
  */
public class PossibleReaction  {

    private static int totalid=0;
    private int id;
    private ReactionType type=null;

    private Vector<VariableSet> localVariableSets=new Vector<VariableSet>(); // one set for each variable
    private Vector<Range> localRanges=new Vector<Range>(); //one range for each parameter

    
    public PossibleReaction(){
	this.id=totalid++;
    }

   /** Read range from input file in form of a buffered reader
     * @param commands vector of strings on first line to read
     * @param in buffered reader
     * @param line_no current line in the buffered reader (i.e., problem file)
     * @param debug debug level
     * @param variables vector of variables
     * @param reactionTypes vector of possible reaction types
     * @param variableSets vector of variable sets
     * @param ranges vector of possible ranges
     */
     public int read(String[] commands,
		     BufferedReader in,
		     int line_no,
		     int debug,
		     Vector<Variable> variables,
		     Vector<ReactionType> reactionTypes,
		     Vector<VariableSet> variableSets,
		     Vector<Range> ranges){
	
	id = Misc.getInteger(commands[0],line_no);
	if (id<=0) {
	    Debug.error(this,"read","id="+id+" should be positive on line="+line_no);
	    System.exit(0);
	}
	if (id!=totalid && id==1) {
	    totalid=1; // start a new variable
	}
	if (id!=totalid) {
	    Debug.error(this,"read","reaction_"+id+" should be reaction_"+totalid+" on line="+line_no);
	    System.exit(0);
	}
	Variable variable=null;
	
	if (commands[1].equals("of") && commands[2].startsWith("variable_")){
	    int ivar=Misc.getInteger(commands[2],line_no);
	    if (ivar>=1 && ivar<=variables.size()){
		variable=variables.get(ivar-1); // compensating for index starting at 0
	    }
	    else{
		Debug.error(this,"read","Illegal variable index: "+commands[2]+", on line="+line_no);
		System.exit(0); 
	    }
	}
	else {
	    Debug.error(this,"read","Unexpected command: "+commands[1]+" "+
			commands[2]+", on line="+line_no);
	    System.exit(0); 
	}



	try {
	    for (;;) {
		String s= in.readLine();
		line_no++;
		commands = Misc.split(s,line_no,debug);
		if (commands==null){
		    if (debug>0) Debug.print(this,"read","empty line");
		    continue;
		}
		else if (commands[0].equals("has") && commands[1].equals("type") &&
			 commands[2].equals("=")) {
		    for (int i=0;i<reactionTypes.size();i++){
			if (commands[3].equals(reactionTypes.get(i).getName())){
			    type=reactionTypes.get(i);
			}
		    }
		    if (type==null) {
			Debug.error(this,"read","Unexpected type: "+commands[3]+", on line="+line_no);
			System.exit(0); 
		    }
		}
		else if (commands[0].equals("has") && commands[1].equals("spaceOfVariable") &&
			 commands[3].equals("=")) {
		    if (type==null) {
			Debug.error(this,"read","Reaction type must be set before "+
				    "assigning spaceOfVariable, on line="+line_no);
			System.exit(0); 
		    }
		    if (!commands[2].equals(type.getLocalVariableName(localVariableSets.size()))){
			Debug.error(this,"read","Unexpected : "+commands[2]+", expected "+
				    type.getLocalVariableName(localVariableSets.size())+
				    ", on line="+line_no);
			System.exit(0); 
		    }
		    if (commands[4].startsWith("memberOfSet_")){
			int i=Misc.getInteger(commands[4],line_no);
			if (i>=1 && i<=variableSets.size()){
			    localVariableSets.add(variableSets.get(i-1)); 
			    // compensating for index starting at 0
			}
			else {
			    Debug.error(this,"read","Illegal integer id: "+commands[4]+
					", on line="+line_no);
			    System.exit(0); 
			}
		    }
		    else if (commands[4].startsWith("variable_")){
			int i=Misc.getInteger(commands[4],line_no);
			if (i>=1 && i<=variables.size()){
			    localVariableSets.add(new VariableSet(variables.get(i-1))); 
			    // compensating for index starting at 0
			}
			else {
			    Debug.error(this,"read","Illegal integer id: "+commands[4]+
					", on line="+line_no);
			    System.exit(0); 
			}
		    }
		    else if (commands[4].equals("dependentVariable")){
			Vector<Variable> members=new Vector<Variable>();
			for (int i=0;i<variables.size();i++) {
			    if (variables.get(i).isDependent()) {
				members.add(variables.get(i));
			    }
			}
			localVariableSets.add(new VariableSet(members)); 
		    }
		    else {
			Debug.error(this,"read","Unexpected varSet: "+commands[4]+", on line="+line_no);
			System.exit(0); 
		    }
		}
		else if (commands[0].equals("has") && commands[1].equals("rangeOfParameter") &&
			 commands[3].equals("=")) {
		    if (type==null) {
			Debug.error(this,"read","Reaction type must be set before "+
				    "assigning spaceOfVariable, on line="+line_no);
			System.exit(0); 
		    }
		    if (!commands[2].equals(type.getLocalParameterName(localRanges.size()))){
			Debug.error(this,"read","Unexpected : "+commands[2]+", expected "+
				    type.getLocalParameterName(localRanges.size())+
				    ", on line="+line_no);
			System.exit(0); 
		    }
		    if (commands[4].startsWith("range_")){
			int i=Misc.getInteger(commands[4],line_no);
			if (i>=1 && i<=ranges.size()){
			    localRanges.add(ranges.get(i-1)); 
			    // compensating for index starting at 0
			}
			else {
			    Debug.error(this,"read","Illegal integer id: "+commands[4]+
					", on line="+line_no);
			    System.exit(0); 
			}
		    }
		    else {
			Debug.error(this,"read","Unexpected range: "+commands[4]+", on line="+line_no);
			System.exit(0); 
		    }

		    
		}
		else {
		    Debug.error(this,"read","Unexpected command: "+commands[0]+", on line="+line_no);
		    System.exit(0); 
		}
		if (localRanges.size()==type.getNumberOfLocalParameters() && 
		    localVariableSets.size()==type.getNumberOfLocalVariables()) break;
	    }
	}
	catch (IOException ioe) {
	    Debug.error(this,"read","IOExpection on line="+line_no);
	    System.exit(0); 
	}

	// Add this possible reaction to variable
	variable.addPossibleReaction(this);
	return line_no;

     }

    /** Get type of the reaction
     * @return the reaction type
     */
    public ReactionType getReactionType(){
	return type;
    }
    
    /** Set type of the reaction
     * @param r the reaction type
     */
    public void setReactionType(ReactionType r){
	type=r;
    }

    /** Get the locally defined variable set i for this possible reaction
     * @param i the index of the variable set
     * @return the variable set 
     */
    public VariableSet getLocalVariableSet(int i){
	return localVariableSets.get(i);
    }

    /** Set the locally defined variable set i for this possible reaction
     * @param i the index of the variable set
     * @param vs the variable set 
     */
    public void setLocalVariableSet(int i, 
				    VariableSet vs){
	localVariableSets.add(i,vs);
    }

    /** Get the number of locally defined variable sets for this possible reaction
     * @return the number of sets
     */
    public int getNumberOfLocalVariableSets(){
	return localVariableSets.size();
    }

    /** Get the locally defined range i for this possible reaction
     * @param i the index of the range
     * @return the range
     */
    public Range getLocalRange(int i){
	return localRanges.get(i);
    }

    /** Set the locally defined range i for this possible reaction
     * @param i the index of the range
     * @param r the range
     */
    public void setLocalRange(int i, Range r){
	localRanges.add(i,r);
    }

    /** Get the number of locally defined ranges for this possible reaction
     * @return the number of ranges 
     */
    public int getNumberOfLocalRanges(){
	return localRanges.size();
    }

    /** Get string representation
     * @return the string
     */ 
    public String toString(){
	String s="Possible reaction "+id+" "+type.getName()+"\n";
	for (int i=0;i<localVariableSets.size();i++)
	    s+=localVariableSets.get(i).toString();
	for (int i=0;i<localRanges.size();i++)
	    s+=localRanges.get(i).toString();	
	return s;
    }
    
    //public static void main(String[] argv) {
    //}
  
}
