/** 
    ODEidon, a program for structural identification of ODEs
    Copyright (C) 2011 Peter Gennemark, Dag Wedelin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Peter Gennemark, Dag Wedelin
    www.odeidentification.org

*/


package Java.problem;

import java.io.*;
import java.util.*;
import Java.debug.*;
import java.lang.Integer;
import Java.misc.*;



/**
 * Store and manage a problem
 * @author Peter Gennemark
 * @version     1.0                   
 * @since       1.0
 */
public class Problem {
    
    // BEGIN CONSTANTS
    private static int DEFAULT_maxpts=201; // max sampling points 
    
    // random starts in rough search (derivative method) and simulate single search
    private static int[] DEFAULT_RAND_PERFECT = {30,1}; 
    private static int[] DEFAULT_RAND_NOISE = {60,1}; 
    private static int[] SS_DEFAULT_RAND_PERFECT = {4,1};
    private static int[] SS_DEFAULT_RAND_NOISE = {8,1};
    
    // Smoothing parameter
    private static double DEFAULT_smoothingParameter_NOISE=0.2; 
    private static double DEFAULT_smoothingParameter_PERFECT=Math.pow(10,-8); 
    
    // random seed
    private static int DEFAULT_SEED=1; 



    // Max number of parameter in an ODE, used in code_rk_estvar.f when calling dn2gb
    private static int MAX_PARAMETERS_PER_ODE=25; 

    // Max number of terms in model space of S-system
    private static int GMA_MAX=6; 

    // END CONSTANTS
    
    
    
    private String name;
    private String type;
    private Vector<Experiment> experiments=new Vector<Experiment>();
    private Vector<ReactionType> reactionTypes=new Vector<ReactionType>();
    private Vector<Variable> variables=new Vector<Variable>();
    private Vector<Range> ranges=new Vector<Range>();
    private Vector<VariableSet> variableSets=new Vector<VariableSet>();
    private ErrorFunction errorFunction;
    private int[] RAND=new int[2];
    private double smoothingParameter;
    private int SEED;


    // S-system, 
    private double[][] alpha=null; // [#vars][initial value, lower bound, upper bound]
    private double[][] beta=null;
    private double[][][] g=null; // [#vars][#vars][initial value, lower bound, upper bound]
    private double[][][] h=null;
    private boolean[][] gnonzero=null;
    private boolean[][] hnonzero=null;

    public Problem() {
	
    }
 


 

    /** Read the problem
     * @param file to be read
     * @param debug debug level
     */
    public void read(File file, 
		     int debug) {
      
	int line_no=0;
	Debug.print(this,"read","File: " + file);
	try {
	    BufferedReader in = new BufferedReader(new FileReader(file));
	    String line="";
       


	    //  Default parameters (noise)
	    boolean perfectData=false;
	    RAND= DEFAULT_RAND_NOISE;
	    smoothingParameter=DEFAULT_smoothingParameter_NOISE;
	    SEED=DEFAULT_SEED;
      
         
      
      
	    for (;;) { 

		if (!in.ready()) break;
		String cline=in.readLine();
		line_no++;//Debug.print(this,"read",line_no);

	  
		if (debug>0) Debug.print(this,"read",cline);



		String[] commands = Misc.split(cline,line_no,debug);
		if (commands==null) {
		    if (debug>0) Debug.print(this,"read","empty line");
		    continue;
		}
	      
	      
		//  ======================================================== // comment
		if (commands[0].startsWith("//")) {
		    if (debug>0) Debug.print(this,"read","comment");
		}

	 

		//  ======================================================== format
		else if (commands[0].equals("format")) {
		    if (debug>0) Debug.print(this,"read","format");
		}

	 
		//  ======================================================== begin
		else if (commands[0].equals("begin")) {
		    if (debug>0) Debug.print(this,"read","begin");
		    if (commands[1].equals("problem")) {
			name=commands[2];
		    }
		    else{
			Debug.error(this,"read","expected problem found "+commands[1]+" on line="+line_no);
			System.exit(0);
		    }
		}


		//  ======================================================== type
		else if (commands[0].equals("type")) {
		    if (debug>0) Debug.print(this,"read","type");  
		    if (commands[2].equals("reactionKinetics")) {
			type=commands[2];
		    }
		    else if (commands[2].equals("SSystem")) {
			type=commands[2];
			RAND= SS_DEFAULT_RAND_NOISE;
		    }
		    else{
			Debug.error(this,"read","expected type to be reactionKinetics or SSystem, found "+commands[2]+" on line="+line_no);
			System.exit(0);
		    }
		}


		//  ======================================================== date
		else if (commands[0].equals("date")) {
		    if (debug>0) Debug.print(this,"read","date");
		}


		//  ======================================================== url 
		else if (commands[0].equals("url")) {
		    if (debug>0) Debug.print(this,"read","url");
		}

		//  ======================================================== reaction_
		else if (commands[0].startsWith("reaction_")) {
		    if (debug>0) Debug.print(this,"read","reaction_");


		    if (commands.length>1 && 
			commands[1].equals("of") && 
			commands[2].startsWith("variable_")) {
			Reaction r=new Reaction();
			line_no=r.read(commands,in,line_no,debug,variables,
				       reactionTypes,ranges); //also add to variable
			if (debug>0) Debug.print(this,"read",r.toString());
		    }
		    else {
			ReactionType r=new ReactionType();
			line_no=r.read(commands,in,line_no,debug);
			if (debug>0) Debug.print(this,"read",r.toString());
			reactionTypes.add(r);
		    }
		}

		//  ======================================================== variable_	has
		else if (commands[0].startsWith("variable_") && commands[1].equals("has")) {
		    if (debug>0) Debug.print(this,"read","variable_");
		    Variable v=new Variable();
		    line_no=v.read(commands,in,line_no,debug);
		    if (debug>0) Debug.print(this,"read",v.toString());
		    variables.add(v);
		}

		//  ======================================================== range__	 
		else if (commands[0].startsWith("range_")) {
		    if (debug>0) Debug.print(this,"read","range_");
		    Range r=new Range();
		    line_no=r.read(commands,in,line_no,debug);
		    if (debug>0) Debug.print(this,"read",r.toString());
		    ranges.add(r);
		}

		//  ======================================================== variableSet_	 
		else if (commands[0].startsWith("memberOfSet_")) {
		    if (debug>0) Debug.print(this,"read","memberOfSet_");
		    VariableSet r=new VariableSet();
		    line_no=r.read(commands,in,line_no,debug,variables);
		    if (debug>0) Debug.print(this,"read",r.toString());
		    variableSets.add(r);
		}


		//  ======================================================== possibleReaction_	 
		else if (commands[0].startsWith("possibleReaction_")) {
		    if (debug>0) Debug.print(this,"read","possibleReaction_");
		    PossibleReaction r=new PossibleReaction();
		    line_no=r.read(commands,in,line_no,debug,variables,
				   reactionTypes,variableSets,ranges); //also add to variable
		    if (debug>0) Debug.print(this,"read",r.toString());
		}


		//  ======================================================== errorFunction_
		else if (commands[0].startsWith("errorFunction")) {
		    if (debug>0) Debug.print(this,"read","errorFunction");
		    errorFunction=new ErrorFunction();
		    line_no=errorFunction.read(commands,in,line_no,debug);
		    if (debug>0) Debug.print(this,"read",errorFunction.toString());
	     
		}


		//  ======================================================== experiment_
		else if (commands[0].startsWith("experiment_")) {
		    if (debug>0) Debug.print(this,"read","experiment");
		    Experiment r=new Experiment(variables.size(),line_no);
		    line_no=r.read(commands,in,line_no,debug);
		    if (debug>0) Debug.print(this,"read",r.toString());
		    experiments.add(r);  
		    if (r.hasPerfectData()) {  // NEW DEFAULTS
			perfectData=false;
			RAND=DEFAULT_RAND_PERFECT;
			if (isSSystem()) RAND=SS_DEFAULT_RAND_PERFECT;
			smoothingParameter=DEFAULT_smoothingParameter_PERFECT;
		    }
		}


		//  ======================================================== sample_	 
		else if (commands[0].startsWith("sample_")) {
		    if (debug>0) Debug.print(this,"read","sample");
		    Sample r=new Sample();
		    line_no=r.read(commands,in,line_no,debug,variables,experiments); //also add to experiment
		    if (debug>0) Debug.print(this,"read",r.toString());
	      
	      
		}
	 

		//  ======================================================== variable_	of (initial bounds)
		else if (commands[0].startsWith("variable_") && commands[1].equals("of")) {
		    if (debug>0) Debug.print(this,"read","variable_ of");
		    InitialBound v=new InitialBound();
		    line_no=v.read(commands,in,line_no,debug,variables,experiments); 
		    // adds itself to experiment
		    if (debug>0) Debug.print(this,"read",v.toString());
	      
		}

	    
		//  ======================================================== randomSeed
		else if (commands[0].equals("randomSeed")) {
		    if (debug>0) Debug.print(this,"read","randomSeed"); 
		    if (commands[1].equals("=")) {
			SEED=Misc.getInteger(commands[2],line_no);
		    }
		    else{
			Debug.error(this,"read","expected = found "+commands[1]+" on line="+line_no);
			System.exit(0);
		    }
		}

	 
		//  ======================================================== smoothingParameter
		else if (commands[0].equals("smoothingParameter")) {
		    if (debug>0) Debug.print(this,"read","smoothingParameter"); 
		    if (commands[1].equals("=")) {
			smoothingParameter=Misc.getDouble(commands[2],line_no);
		    }
		    else{
			Debug.error(this,"read","expected = found "+commands[1]+" on line="+line_no);
			System.exit(0);
		    }
		}


		//  ======================================================== randomStarts_
		else if (commands[0].equals("randomStarts_")) {
		    if (debug>0) Debug.print(this,"read","randomStarts_"); 
		    if (commands[1].equals("=")) {
			RAND[0]= Misc.getInteger(commands[2],line_no);
			RAND[1]= Misc.getInteger(commands[3],line_no);
		    }
		    else{
			Debug.error(this,"read","expected = found "+commands[1]+" on line="+line_no);
			System.exit(0);
		    }
		}



		//  ======================================================== end
		else if (commands[0].equals("end")) {
		    if (debug>0) Debug.print(this,"read","end");
		    //break; // don not break here, random seed may be defined manually after end
		}



		//  ======================================================== S-system, alpha defaultInitialValue
		else if (commands[0].equals("alpha") && 
			 commands[1].equals("has") &&  
			 commands[2].equals("defaultInitialValue")) {
		    if (debug>0) Debug.print(this,"read","alpha defaultInitialValue");
		    if (alpha==null) alpha=new double[getNumberOfVariables()][3];
		    alpha[0][0]=Misc.getDouble(commands[4],line_no);
		    for (int i=1;i<alpha.length;i++) alpha[i][0]=alpha[0][0];
		}
		//  ======================================================== S-system, alpha defaultLowerBound
		else if (commands[0].equals("alpha") && 
			 commands[1].equals("has") &&  
			 commands[2].equals("defaultLowerBound")) {
		    if (debug>0) Debug.print(this,"read","alpha defaultLowerBound");
		    if (alpha==null) alpha=new double[getNumberOfVariables()][3];
		    alpha[0][1]=Misc.getDouble(commands[4],line_no);
		    for (int i=1;i<alpha.length;i++) alpha[i][1]=alpha[0][1];
		}

		//  ======================================================== S-system, alpha defaultUpperBound
		else if (commands[0].equals("alpha") && 
			 commands[1].equals("has") &&  
			 commands[2].equals("defaultUpperBound")) {
		    if (debug>0) Debug.print(this,"read","alpha defaultUpperBound");
		    if (alpha==null) alpha=new double[getNumberOfVariables()][3];
		    alpha[0][2]=Misc.getDouble(commands[4],line_no);
		    for (int i=1;i<alpha.length;i++) alpha[i][2]=alpha[0][2];
		}

		//  ======================================================== S-system, beta defaultInitialValue
		else if (commands[0].equals("beta") && 
			 commands[1].equals("has") &&  
			 commands[2].equals("defaultInitialValue")) {
		    if (debug>0) Debug.print(this,"read","beta defaultInitialValue");
		    if (beta==null) beta=new double[getNumberOfVariables()][3];
		    beta[0][0]=Misc.getDouble(commands[4],line_no);
		    for (int i=1;i<beta.length;i++) beta[i][0]=beta[0][0];
		}
		//  ======================================================== S-system, beta defaultLowerBound
		else if (commands[0].equals("beta") && 
			 commands[1].equals("has") &&  
			 commands[2].equals("defaultLowerBound")) {
		    if (debug>0) Debug.print(this,"read","beta defaultLowerBound");
		    if (beta==null) beta=new double[getNumberOfVariables()][3];
		    beta[0][1]=Misc.getDouble(commands[4],line_no);
		    for (int i=1;i<beta.length;i++) beta[i][1]=beta[0][1];
		}

		//  ======================================================== S-system, beta defaultUpperBound
		else if (commands[0].equals("beta") && 
			 commands[1].equals("has") &&  
			 commands[2].equals("defaultUpperBound")) {
		    if (debug>0) Debug.print(this,"read","beta defaultUpperBound");
		    if (beta==null) beta=new double[getNumberOfVariables()][3];
		    beta[0][2]=Misc.getDouble(commands[4],line_no);
		    for (int i=1;i<beta.length;i++) beta[i][2]=beta[0][2];
		}


		//  ======================================================== S-system, g defaultInitialValue
		else if (commands[0].equals("g") && 
			 commands[1].equals("has") &&  
			 commands[2].equals("defaultInitialValue")) {
		    if (debug>0) Debug.print(this,"read","g defaultInitialValue");
		    if (g==null) g=new double[getNumberOfVariables()][getNumberOfVariables()][3];
		    double d=Misc.getDouble(commands[4],line_no);
		    for (int i=0;i<getNumberOfVariables();i++)
			for (int j=0;j<getNumberOfVariables();j++) 
			    g[i][j][0]=d;
		}
		//  ======================================================== S-system, g defaultLowerBound
		else if (commands[0].equals("g") && 
			 commands[1].equals("has") &&  
			 commands[2].equals("defaultLowerBound")) {
		    if (debug>0) Debug.print(this,"read","g defaultLowerBound");
		    if (g==null) g=new double[getNumberOfVariables()][getNumberOfVariables()][3];
		    double d=Misc.getDouble(commands[4],line_no);
		    for (int i=0;i<getNumberOfVariables();i++)
			for (int j=0;j<getNumberOfVariables();j++) 
			    g[i][j][1]=d;
		}
		//  ======================================================== S-system, g defaultUpperBound
		else if (commands[0].equals("g") && 
			 commands[1].equals("has") &&  
			 commands[2].equals("defaultUpperBound")) {
		    if (debug>0) Debug.print(this,"read","g defaultUpperBound");
		    if (g==null) g=new double[getNumberOfVariables()][getNumberOfVariables()][3];
		    double d=Misc.getDouble(commands[4],line_no);
		    for (int i=0;i<getNumberOfVariables();i++)
			for (int j=0;j<getNumberOfVariables();j++) 
			    g[i][j][2]=d;
		}
		//  ======================================================== S-system, h defaultInitialValue
		else if (commands[0].equals("h") && 
			 commands[1].equals("has") &&  
			 commands[2].equals("defaultInitialValue")) {
		    if (debug>0) Debug.print(this,"read","h defaultInitialValue");
		    if (h==null) h=new double[getNumberOfVariables()][getNumberOfVariables()][3];
		    double d=Misc.getDouble(commands[4],line_no);
		    for (int i=0;i<getNumberOfVariables();i++)
			for (int j=0;j<getNumberOfVariables();j++) 
			    h[i][j][0]=d;
		}
		//  ======================================================== S-system, h defaultLowerBound
		else if (commands[0].equals("h") && 
			 commands[1].equals("has") &&  
			 commands[2].equals("defaultLowerBound")) {
		    if (debug>0) Debug.print(this,"read","h defaultLowerBound");
		    if (h==null) h=new double[getNumberOfVariables()][getNumberOfVariables()][3];
		    double d=Misc.getDouble(commands[4],line_no);
		    for (int i=0;i<getNumberOfVariables();i++)
			for (int j=0;j<getNumberOfVariables();j++) 
			    h[i][j][1]=d;
		}
		//  ======================================================== S-system, h defaultUpperBound
		else if (commands[0].equals("h") && 
			 commands[1].equals("has") &&  
			 commands[2].equals("defaultUpperBound")) {
		    if (debug>0) Debug.print(this,"read","h defaultUpperBound");
		    if (h==null) h=new double[getNumberOfVariables()][getNumberOfVariables()][3];
		    double d=Misc.getDouble(commands[4],line_no);
		    for (int i=0;i<getNumberOfVariables();i++)
			for (int j=0;j<getNumberOfVariables();j++) 
			    h[i][j][2]=d;
		}




		//  ======================================================== S-system, alpha initialValue
		else if (commands[0].startsWith("alpha_")&&
			 commands[1].equals("has") && 
			 commands[2].equals("initialValue")) {
		    if (debug>0) Debug.print(this,"read","alpha_ initialValue");
		    if (alpha==null) alpha=new double[getNumberOfVariables()][3];
		    int index=Misc.getInteger(commands[0],line_no);
		    alpha[index][0]=Misc.getDouble(commands[4],line_no);
		}
		//  ======================================================== S-system, alpha lowerBound
		else if (commands[0].startsWith("alpha_")&&
			 commands[1].equals("has") && 
			 commands[2].equals("lowerBound")) {
		    if (debug>0) Debug.print(this,"read","alpha_ lowerBound");
		    if (alpha==null) alpha=new double[getNumberOfVariables()][3];
		    int index=Misc.getInteger(commands[0],line_no);
		    alpha[index][1]=Misc.getDouble(commands[4],line_no);
		}
		//  ======================================================== S-system, alpha upperBound
		else if (commands[0].startsWith("alpha_")&&
			 commands[1].equals("has") && 
			 commands[2].equals("upperBound")) {
		    if (debug>0) Debug.print(this,"read","alpha_ upperBound");
		    if (alpha==null) alpha=new double[getNumberOfVariables()][3];
		    int index=Misc.getInteger(commands[0],line_no);
		    alpha[index][2]=Misc.getDouble(commands[4],line_no);
		}


		//  ======================================================== S-system, beta initialValue
		else if (commands[0].startsWith("beta_")&&
			 commands[1].equals("has") && 
			 commands[2].equals("initialValue")) {
		    if (debug>0) Debug.print(this,"read","beta_ initialValue");
		    if (beta==null) beta=new double[getNumberOfVariables()][3];
		    int index=Misc.getInteger(commands[0],line_no);
		    beta[index][0]=Misc.getDouble(commands[4],line_no);
		}
		//  ======================================================== S-system, beta lowerBound
		else if (commands[0].startsWith("beta_")&&
			 commands[1].equals("has") && 
			 commands[2].equals("lowerBound")) {
		    if (debug>0) Debug.print(this,"read","beta_ lowerBound");
		    if (beta==null) beta=new double[getNumberOfVariables()][3];
		    int index=Misc.getInteger(commands[0],line_no);
		    beta[index][1]=Misc.getDouble(commands[4],line_no);
		}
		//  ======================================================== S-system, beta upperBound
		else if (commands[0].startsWith("beta_")&&
			 commands[1].equals("has") && 
			 commands[2].equals("upperBound")) {
		    if (debug>0) Debug.print(this,"read","beta_ upperBound");
		    if (beta==null) beta=new double[getNumberOfVariables()][3];
		    int index=Misc.getInteger(commands[0],line_no);
		    beta[index][2]=Misc.getDouble(commands[4],line_no);
		}




		//  ======================================================== S-system, g nonZero
		else if (commands[0].startsWith("g_")&&
			 commands[1].equals("is") && 
			 commands[2].equals("nonZero")) {
		    if (debug>0) Debug.print(this,"read","g_ is nonZero");
		    if (gnonzero==null) gnonzero=new boolean[getNumberOfVariables()][getNumberOfVariables()];
		    int[] i=Misc.getIntegers(commands[0],line_no);
		    gnonzero[i[0]-1][i[1]-1]=true;
		}
		//  ======================================================== S-system, g initialValue
		else if (commands[0].startsWith("g_")&&
			 commands[1].equals("has") && 
			 commands[2].equals("initialValue")) {
		    if (debug>0) Debug.print(this,"read","g_ initialValue");
		    if (g==null) g=new double[getNumberOfVariables()][getNumberOfVariables()][3];
		    int[] i=Misc.getIntegers(commands[0],line_no);
		    g[i[0]-1][i[1]-1][0]=Misc.getDouble(commands[4],line_no);
		}
		//  ======================================================== S-system, g lowerBound
		else if (commands[0].startsWith("g_")&&
			 commands[1].equals("has") && 
			 commands[2].equals("lowerBound")) {
		    if (debug>0) Debug.print(this,"read","g_ lowerBound");
		    if (g==null) g=new double[getNumberOfVariables()][getNumberOfVariables()][3];
		    int[] i=Misc.getIntegers(commands[0],line_no);
		    g[i[0]-1][i[1]-1][1]=Misc.getDouble(commands[4],line_no);
		}

		//  ======================================================== S-system, g upperBound
		else if (commands[0].startsWith("g_")&&
			 commands[1].equals("has") && 
			 commands[2].equals("upperBound")) {
		    if (debug>0) Debug.print(this,"read","g_ upperBound");
		    if (g==null) g=new double[getNumberOfVariables()][getNumberOfVariables()][3];
		    int[] i=Misc.getIntegers(commands[0],line_no);
		    g[i[0]-1][i[1]-1][2]=Misc.getDouble(commands[4],line_no);
		}



		//  ======================================================== S-system, h nonZero
		else if (commands[0].startsWith("h_")&&
			 commands[1].equals("is") && 
			 commands[2].equals("nonZero")) {
		    if (debug>0) Debug.print(this,"read","h_ is nonZero");
		    if (hnonzero==null) hnonzero=new boolean[getNumberOfVariables()][getNumberOfVariables()];
		    int[] i=Misc.getIntegers(commands[0],line_no);
		    hnonzero[i[0]-1][i[1]-1]=true;
		}
		//  ======================================================== S-system, h initialValue
		else if (commands[0].startsWith("h_")&&
			 commands[1].equals("has") && 
			 commands[2].equals("initialValue")) {
		    if (debug>0) Debug.print(this,"read","h_ initialValue");
		    if (h==null) h=new double[getNumberOfVariables()][getNumberOfVariables()][3];
		    int[] i=Misc.getIntegers(commands[0],line_no);
		    h[i[0]-1][i[1]-1][0]=Misc.getDouble(commands[4],line_no);
		}
		//  ======================================================== S-system, h lowerBound
		else if (commands[0].startsWith("h_")&&
			 commands[1].equals("has") && 
			 commands[2].equals("lowerBound")) {
		    if (debug>0) Debug.print(this,"read","h_ lowerBound");
		    if (h==null) h=new double[getNumberOfVariables()][getNumberOfVariables()][3];
		    int[] i=Misc.getIntegers(commands[0],line_no);
		    h[i[0]-1][i[1]-1][1]=Misc.getDouble(commands[4],line_no);
		}

		//  ======================================================== S-system, h upperBound
		else if (commands[0].startsWith("h_")&&
			 commands[1].equals("has") && 
			 commands[2].equals("upperBound")) {
		    if (debug>0) Debug.print(this,"read","h_ upperBound");
		    if (h==null) h=new double[getNumberOfVariables()][getNumberOfVariables()][3];
		    int[] i=Misc.getIntegers(commands[0],line_no);
		    h[i[0]-1][i[1]-1][2]=Misc.getDouble(commands[4],line_no);
		}


		else{
		    Debug.error(this,"read","Illegal command: "+cline+", line: " + line_no);
		    System.exit(0);
		}



	    } //main loop

	    if (isSSystem() && gnonzero==null) gnonzero=new boolean[getNumberOfVariables()][getNumberOfVariables()];
	    if (isSSystem() && hnonzero==null) hnonzero=new boolean[getNumberOfVariables()][getNumberOfVariables()];
	    //Debug.print(this,"read",experiments.get(0).toString());
	    //Debug.print(this,"read",experiments.get(1).toString());
	    //System.out.println(variables.get(1).toString());
	    //Debug.print(this,"read",variables.get(3).toString());
	    //Debug.print(this,"read",variables.get(4).toString());


	    //Debug.print(this,"read",reactionTypes.get(0).toString());
	    //Debug.print(this,"read",reactionTypes.get(2).toString());



	    if (isSSystem()) {
		adaptSSystem();
	    }


	}catch(IOException ioe) {    
	    String errorMsg="Error reading file " + file + ", line number: " + line_no;   
	    Debug.print(this,"read",errorMsg); 
	    Debug.print(this,"read",ioe.toString()); 
	    //System.exit(0);	
	} 











    }//end of read



     /** String value of experiments, calling toString in Experiment
      * @return the string value of the experiments
      */  
    public String toString(){
	String ret="";
	for (int i=0;i<experiments.size();i++) {
	    ret=ret+ ((Experiment)experiments.get(i)).toString()+ "\n";
	}
	ret=ret+"\n";
	return ret;
    }
    
    /** Get the number of experiments
     * @return the number
     */ 
    public int getNumberOfExperiments(){
	return experiments.size();
    }
    
    /** Get the number of reaction types
     * @return the number
     */ 
    public int getNumberOfReactionTypes(){
	return reactionTypes.size();
    }
    
    /** Get the number of variables
     * @return the number
     */ 
    public int getNumberOfVariables(){
	return variables.size();
    }
    
    /** Get the number of dependent variables
     * @return the number
     */ 
    public int getNumberOfDependentVariables(){
	int c=0;
	for (int i=0;i<getNumberOfVariables();i++){
	    if (variables.get(i).isDependent()) c++;
	}
	return c;
    }
    
    /** Get the number of ranges
     * @return the number
     */ 
    public int getNumberOfRanges(){
	return ranges.size();
    }
    
    /** Get experiment i
     * @param i the index
     * @return the experiment
     */ 
    public Experiment getExperiment(int i){
	return experiments.get(i);
    }
    
    /** Get reaction type i
     * @param i the index
     * @return the reaction type
     */ 
    public ReactionType getReactionType(int i){
	return reactionTypes.get(i);
    }
    
    /** Get all reaction types in a vector
     * @return the reaction types
     */ 
    public Vector<ReactionType> getReactionTypes(){
	return reactionTypes;
    }
    
    /** Get all variables in a vector
     * @return the variables
     */ 
    public Vector<Variable> getVariables(){
	return variables;
    }
    
    /** Get reaction type named nameOfType
     * @param nameOfType the name
     * @return the reaction type
     */ 
    public ReactionType getReactionType(String nameOfType){
	for (int i=0;i<getNumberOfReactionTypes();i++){
	    if (getReactionType(i).getName().equals(nameOfType)){
		return getReactionType(i);
	    }
	}
	return null;
    }
    
    /** Get variable i
     * @param i the index
     * @return the variable
     */ 
    public Variable getVariable(int i){
	return variables.get(i);
    }
    
    /** Get range i
     * @param i the index
     * @return the range
     */ 
    public Range getRange(int i){
	return ranges.get(i);
    }
    
    /** Get variable set i
     * @param i the index
     * @return the variable set
     */ 
    public VariableSet getVariableSet(int i){
	return variableSets.get(i);
    }
    
    /** Get maximum number of sampling points
     * @return the number
     */ 
    public int getMaxpts(){
	return DEFAULT_maxpts;
    }  
    
    /** Get random seed
     * @return the seed
     */
    public int getSEED(){
	return SEED;
    }
    
    /** Get smoothing parameter
     * @return the parameter
     */
    public double getSmoothingParameter(){
	return smoothingParameter;
    }
    
    /** Get number of random starts in local parameter estimation 
     * @return the number
     */
    public int getRAND1(){
	return RAND[0];
    }
    
    /** Get number of random starts in global parameter estimation 
     * @return the number
     */
    public int getRAND2(){
	return RAND[1];
    }
    
    /** Get problem name
     * @return the name
     */
    public String getName(){
	return name;
    }
    
    /** Get problem type
     * @return the type
     */
    public String getType(){
	return type;
    }
    
    /** Is the problem of S-system type
     * @return true if it is an S-system, ow false
     */
    public boolean isSSystem(){
	return type.equals("SSystem");
    }
    
    /** Get error function
     * @return the error function
     */
    public  ErrorFunction getErrorFunction() {
	return errorFunction;
    }
    
    /** Get maximum number of samples in the experiments
     * @return the number
     */
    public int get_maxpts(){
	int maxpts=0;
	for (int i=0;i<getNumberOfExperiments();i++){
	    if (getExperiment(i).getNumberOfSamples()>maxpts) 
		maxpts=getExperiment(i).getNumberOfSamples();
	}
	return maxpts;
    }
    
    /** Get maximum number of variables in a reaction type
     * @return the number
     */
    public int get_maxv(){
	int maxv=0;
	for (int i=0;i<getNumberOfReactionTypes();i++){
	    if (getReactionType(i).getNumberOfLocalVariables()>maxv) 
		maxv=getReactionType(i).getNumberOfLocalVariables();
	}
	return maxv;
    }
    
    /** Get maximum number of parameters in a reaction type
     * @return the number
     */
    public int get_maxk(){
	int maxk=0;
	for (int i=0;i<getNumberOfReactionTypes();i++){
	    if (getReactionType(i).getNumberOfLocalParameters()>maxk) 
		maxk=getReactionType(i).getNumberOfLocalParameters();
	}
	return maxk;
    }
    
    /** Get maximum number of parameters per ODE
     * @return the number
     */
    public int getMAX_PARAMETERS_PER_ODE(){
	return MAX_PARAMETERS_PER_ODE;
    }
    
    /** Get maximum number of samples for one variable 
     * @return the number
     */
    public int getMaxNumberOfSamplesPerVariable(){
	int max=0;
	for (int i=0;i<getNumberOfExperiments();i++){
	    max+=getExperiment(i).getNumberOfSamples();
	}
	return max;
    }
    
    /** Get the alpha matrix, relevant if problem is based on an S-system
     * @return the matrix
     */
    public double[][] getAlpha(){
	return alpha;
    }
    
    /** Get the beta matrix, relevant if problem is based on an S-system
     * @return the matrix
     */
    public double[][] getBeta(){
	return beta;
    }
    
    /** Get the g matrix, relevant if problem is based on an S-system
     * @return the matrix
     */
    public double[][][] getG(){
	return g;
    }
    
    /** Get the h matrix, relevant if problem is based on an S-system
     * @return the matrix
     */
    public double[][][] getH(){
	return h;
    }
    
    /** Get information of non-zero elements of the g matrix, 
     * relevant if problem is based on an S-system
     * @return a boolean matrix, true if element is non-zero, ow false
     */
    public boolean[][] getGnonzero(){
	return gnonzero;
    }
    
    /** Get information of non-zero elements of the h matrix, 
     * relevant if problem is based on an S-system
     * @return a boolean matrix, true if element is non-zero, ow false
     */
    public boolean[][] getHnonzero(){
	return hnonzero;
    }
    
    /** Adapt information of S-system to a reaction kinetic problem 
     * relevant if problem is based on an S-system
     */
    public void adaptSSystem(){

	if (alpha==null || beta==null || h==null || g==null) {
	    printSSystem();
	    Debug.error(this,"adaptSSystem","Type is "+type+
			" but at least one of alpha, beta, g and h is undefined (null).");
	    System.exit(0);
	} 

	// Reaction types
	for (int gma=0;gma<=GMA_MAX;gma++){
	    PossibleReaction r=new PossibleReaction();
	    ReactionType rt=new ReactionType();
	    rt.setAsGMAterm(gma);
	    //System.out.print(rt);
	    reactionTypes.add(rt);
	}


	// Is the model space empty? Then a parameter estimation problem
	boolean isParaEst=true;
	for (int i=0;i<getNumberOfVariables();i++){
	    for (int j=0;j<getNumberOfVariables();j++){
		if (g[i][j][0]==0.0 && (g[i][j][1]!=0.0 ||  g[i][j][2]!=0.0)) isParaEst=false;
		if (h[i][j][0]==0.0 && (h[i][j][1]!=0.0 ||  h[i][j][2]!=0.0)) isParaEst=false;
	    }
	}

	// Possible reactions, 
	// Add one dummy possible reaction to each dependent variable
	// If a parameter est problem, no model space
	if (!isParaEst) {
	    for (int i=0;i<getNumberOfVariables();i++){
		if (getVariable(i).isDependent()) {
		    PossibleReaction r=new PossibleReaction();
		    ReactionType rt=new ReactionType();
		    rt.setAsGMAterm(0);
		    r.setReactionType(rt);
		    //VariableSet vs=new VariableSet(variables); // all variables
		    //r.setLocalVariableSet(0,vs);
		    r.setLocalRange(0,new Range(0.0,1.0));
		    getVariable(i).addPossibleReaction(r);
		}
	    }
	}

	// Inital model
	for (int i=0;i<getNumberOfVariables();i++){
	    if (getVariable(i).isDependent()) {
		// alpha and g
		Reaction r=new Reaction();
		r.setLocalRange(0,new Range(alpha[i][1],alpha[i][2],alpha[i][0]));
		int gma=0;
		for (int j=0;j<getNumberOfVariables();j++) {
		    if (g[i][j][0]!=0.0) {
			r.setLocalVariable(gma,getVariable(j));
			r.setLocalRange(gma+1,new Range(g[i][j][1],g[i][j][2],g[i][j][0]));
			gma++;
		    }
		}
		ReactionType rt=new ReactionType();
		rt.setAsGMAterm(gma);
		r.setReactionType(rt);
		getVariable(i).addReaction(r);
		// beta and h
		r=new Reaction();
		r.setLocalRange(0,new Range(-beta[i][2],beta[i][1],-beta[i][0]));
		gma=0;
		for (int j=0;j<getNumberOfVariables();j++) {
		    if (h[i][j][0]!=0.0) {
			r.setLocalVariable(gma,getVariable(j));
			r.setLocalRange(gma+1,new Range(h[i][j][1],h[i][j][2],h[i][j][0]));
			gma++;
		    }
		}
		rt=new ReactionType();
		rt.setAsGMAterm(gma);
		r.setReactionType(rt);
		getVariable(i).addReaction(r);
		//System.out.print(getVariable(i));
	    }
	}

    }
    
    /** Print information of an S-system, relevant if problem is based on an S-system
     */
    public void printSSystem(){
	if (alpha!=null){
	    System.out.println("ALPHA, initial values");
	    for (int i=0;i<alpha.length;i++){
		System.out.print(alpha[i][0]+" ");
	    }
	    System.out.println();
	    System.out.println("ALPHA, lower bounds");
	    for (int i=0;i<alpha.length;i++){
		System.out.print(alpha[i][1]+" ");
	    }
	    System.out.println();
	    System.out.println("ALPHA, upper bounds");
	    for (int i=0;i<alpha.length;i++){
		System.out.print(alpha[i][2]+" ");
	    }
	    System.out.println();
	}
	if (beta!=null){
	    System.out.println("BETA, initial values");
	    for (int i=0;i<beta.length;i++){
		System.out.print(beta[i][0]+" ");
	    }
	    System.out.println();
	    System.out.println("BETA, lower bounds");
	    for (int i=0;i<beta.length;i++){
		System.out.print(beta[i][1]+" ");
	    }
	    System.out.println();
	    System.out.println("BETA, upper bounds");
	    for (int i=0;i<beta.length;i++){
		System.out.print(beta[i][2]+" ");
	    }
	    System.out.println();
	}

	if (g!=null){
	    System.out.println("g, initial values");
	    for (int i=0;i<getNumberOfVariables();i++){
		for (int j=0;j<getNumberOfVariables();j++){
		    System.out.print(g[i][j][0]+" ");
		}
		System.out.println();
	    }
	    System.out.println("g, lower bounds");
	    for (int i=0;i<getNumberOfVariables();i++){
		for (int j=0;j<getNumberOfVariables();j++){
		    System.out.print(g[i][j][1]+" ");
		}
		System.out.println();
	    }
	    System.out.println("g, upper bounds");
	    for (int i=0;i<getNumberOfVariables();i++){
		for (int j=0;j<getNumberOfVariables();j++){
		    System.out.print(g[i][j][2]+" ");
		}
		System.out.println();
	    }
	    System.out.println("g, non-zeros");
	    for (int i=0;i<getNumberOfVariables();i++){
		for (int j=0;j<getNumberOfVariables();j++){
		    if (gnonzero[i][j]) System.out.print(" T  ");
		    else System.out.print("    ");
		}
		System.out.println();
	    }
	}
	if (h!=null){
	    System.out.println("h, initial values");
	    for (int i=0;i<getNumberOfVariables();i++){
		for (int j=0;j<getNumberOfVariables();j++){
		    System.out.print(h[i][j][0]+" ");
		}
		System.out.println();
	    }
	    System.out.println("h, lower bounds");
	    for (int i=0;i<getNumberOfVariables();i++){
		for (int j=0;j<getNumberOfVariables();j++){
		    System.out.print(h[i][j][1]+" ");
		}
		System.out.println();
	    }
	    System.out.println("h, upper bounds");
	    for (int i=0;i<getNumberOfVariables();i++){
		for (int j=0;j<getNumberOfVariables();j++){
		    System.out.print(h[i][j][2]+" ");
		}
		System.out.println();
	    }
	    System.out.println("h, non-zeros");
	    for (int i=0;i<getNumberOfVariables();i++){
		for (int j=0;j<getNumberOfVariables();j++){
		    if (hnonzero[i][j]) System.out.print(" T  ");
		    else System.out.print("    ");
		}
		System.out.println();
	    }
	}
    }
    
  
    // public static void main(String[] argv){  

    
    //   // File system variables
    //   String filetree = System.getProperty("user.dir") + 
    //     System.getProperty("file.separator");
    
    //   // Problem file
    //   File file=new File(filetree+argv[0]);
  
    
    
    //   Problem p=new Problem();
    //   p.read(file,1);    

    //   //System.out.println(p.getSEED());

    // }


}


