/** 
    ODEidon, a program for structural identification of ODEs
    Copyright (C) 2011 Peter Gennemark, Dag Wedelin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Peter Gennemark, Dag Wedelin
             www.odeidentification.org

*/



package Java.problem;

import java.util.*;
import Java.debug.*;
import java.io.*;
import Java.misc.*;

/**
  * Description of a range with lower and upper bounds, and an initial guess<BR>
  * @author Peter Gennemark
  * @version     1.0                   
  * @since       1.0
  */

public class Range  {
    private static int totalid=0;
    private int id;
    private double lowerBound;
    private double upperBound;
    private double initialGuess;
     
    public Range(){
	this.id=totalid++;
    }
    
    /** 
     * @param lb lower bound
     * @param ub upper bound
     */
    public Range(double lb,double ub){
	this.id=totalid++;
	lowerBound=lb;
	upperBound=ub;
	initialGuess=0.0;
    }
    
    /** 
     * @param lb lower bound
     * @param ub upper bound
     * @param ig initial guesss
     */
    public Range(double lb,double ub,double ig){
	this.id=totalid++;
	lowerBound=lb;
	upperBound=ub;
	initialGuess=ig;
    }
    
   /** Read range from input file in form of a buffered reader
     * @param commands vector of strings on first line to read
     * @param in buffered reader
     * @param line_no current line in the buffered reader (i.e., problem file)
     * @param debug debug level
     */
    public int read(String[] commands,
		    BufferedReader in,
		    int line_no,
		    int debug){
	
	id = Misc.getInteger(commands[0],line_no);
	if (id<=0) {
	    Debug.error(this,"read","id="+id+" should be positive on line="+line_no);
	    System.exit(0);
	}
	if (id!=totalid) {
	    Debug.error(this,"read","range_"+id+" should be range_"+totalid+" on line="+line_no);
	    System.exit(0);
	}
	if (commands[1].equals("has") && commands[2].equals("lowerBound") &&
	    commands[3].equals("=")) {
	    lowerBound=Misc.getDouble(commands[4],line_no);
	}
	else {
	    Debug.error(this,"read","Unexpected command: "+commands[1]+" "+
			commands[2]+" "+commands[3]+", on line="+line_no);
	    System.exit(0); 
	}
	if (commands[5].equals("has") && commands[6].equals("upperBound") &&
	    commands[7].equals("=")) {
	    upperBound=Misc.getDouble(commands[8],line_no);
	}
	else {
	    Debug.error(this,"read","Unexpected command: "+commands[5]+" "+
			commands[6]+" "+commands[7]+", on line="+line_no);
	    System.exit(0); 
	}
	initialGuess=lowerBound;
	if (upperBound<0.0) initialGuess=upperBound; 
   	return line_no;
	
    }

    /** Get lower bound of the range
     * @return the lower bound
     */ 
    public double getLowerBound(){
	return lowerBound;
    }

    /** Get upper bound of the range
     * @return the upper bound
     */ 
    public double getUpperBound(){
	return upperBound;
    }

    /** Get initial guess, a value taken from the range
     * @return the guess
     */ 
    public double getInitialGuess(){
	return initialGuess;
    }

    /** Get string representation
     * @return the string
     */ 
    public String toString(){
	String s="Range "+id+": lb: "+lowerBound+" ub: "+upperBound+
	    " initguess: "+initialGuess+"\n";
	return s;
    }

    // public static void main(String[] argv) {
    // }
  
}
