/** 
    ODEidon, a program for structural identification of ODEs
    Copyright (C) 2011 Peter Gennemark, Dag Wedelin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Peter Gennemark, Dag Wedelin
             www.odeidentification.org

*/



package Java.problem;

import java.util.*;
import Java.debug.*;
import java.io.*;
import Java.misc.*;

/**
  * Description of a reaction given in the model<BR> 
  * If an initial model in a problem, then one range for each parameter
  * stored in a Vector called localRanges<BR> 
  * If a solution model, then one value for each parameter
  * stored in a Vector called localParameters
  * @author Peter Gennemark
  * @version     1.0                   
  * @since       1.0
  */
public class Reaction  {


    private static int totalid=0;
    private int id;
    private ReactionType type=null;

    private Vector<Variable> localVariables=new Vector<Variable>(); 

    // If initial model in a problem, then a range
    private Vector<Range> localRanges=new Vector<Range>(); //one range for each parameter

    // If a solution model, then a double
    private Vector<Double> localParameters=new Vector<Double>(); //or one value for each parameter

    
    public Reaction(){
	this.id=totalid++;
    }

    /** Read reaction from input file in form of a buffered reader
     * @param commands vector of strings on first line to read
     * @param in buffered reader
     * @param line_no current line in the buffered reader (i.e., problem file)
     * @param debug debug level
     * @param variables vector of variables
     * @param reactionTypes vector of possible reaction types
     * @param ranges vector of possible ranges
     */
     public int read(String[] commands,
		     BufferedReader in,
		     int line_no,
		     int debug,
		     Vector<Variable> variables,
		     Vector<ReactionType> reactionTypes,
		     Vector<Range> ranges){
	
	id = Misc.getInteger(commands[0],line_no);
	if (id<=0) {
	    Debug.error(this,"read","id="+id+" should be positive on line="+line_no);
	    System.exit(0);
	}
	if (id!=totalid && id==1) {
	    totalid=1; // start a new variable
	}
	if (id!=totalid) {
	    Debug.error(this,"read","reaction_"+id+" should be reaction_"+totalid+" on line="+line_no);
	    System.exit(0);
	}
	Variable variable=null;
	
	if (commands[1].equals("of") && commands[2].startsWith("variable_")){
	    int ivar=Misc.getInteger(commands[2],line_no);
	    if (ivar>=1 && ivar<=variables.size()){
		variable=variables.get(ivar-1); // compensating for index starting at 0
	    }
	    else{
		Debug.error(this,"read","Illegal variable index: "+commands[2]+", on line="+line_no);
		System.exit(0); 
	    }
	}
	else {
	    Debug.error(this,"read","Unexpected command: "+commands[1]+" "+
			commands[2]+", on line="+line_no);
	    System.exit(0); 
	}



	try {
	    for (;;) {
		String s= in.readLine();
		line_no++;
		commands = Misc.split(s,line_no,debug);
		if (commands==null){
		    if (debug>0) Debug.print(this,"read","empty line");
		    continue;
		}
		else if (commands[0].equals("has") && commands[1].equals("type") &&
			 commands[2].equals("=")) {
		    for (int i=0;i<reactionTypes.size();i++){
			if (commands[3].equals(reactionTypes.get(i).getName())){
			    type=reactionTypes.get(i);
			}
		    }
		    if (type==null) {
			Debug.error(this,"read","Unexpected type: "+commands[3]+", on line="+line_no);
			System.exit(0); 
		    }
		}
		else if (commands[0].equals("has") && commands[1].equals("variable") &&
			 commands[3].equals("=")) {
		    if (type==null) {
			Debug.error(this,"read","Reaction type must be set before "+
				    "assigning spaceOfVariable, on line="+line_no);
			System.exit(0); 
		    }
		    if (!commands[2].equals(type.getLocalVariableName(localVariables.size()))){
			Debug.error(this,"read","Unexpected : "+commands[2]+", expected "+
				    type.getLocalVariableName(localVariables.size())+
				    "on line="+line_no);
			System.exit(0); 
		    }
		    if (commands[4].startsWith("variable_")){
			int i=Misc.getInteger(commands[4],line_no);
			if (i>=1 && i<=variables.size()){
			    localVariables.add(variables.get(i-1)); 
			    // compensating for index starting at 0
			}
			else {
			    Debug.error(this,"read","Illegal integer id: "+commands[4]+
					", on line="+line_no);
			    System.exit(0); 
			}
		    }
		    else {
			Debug.error(this,"read","Unexpected variable: "+commands[4]+", on line="+line_no);
			System.exit(0); 
		    }
		}
		else if (commands[0].equals("has") && commands[1].equals("rangeOfParameter") &&
			 commands[3].equals("=")) {
		    if (type==null) {
			Debug.error(this,"read","Reaction type must be set before "+
				    "assigning rangeOfParameter, on line="+line_no);
			System.exit(0); 
		    }
		    if (!commands[2].equals(type.getLocalParameterName(localRanges.size()))){
			Debug.error(this,"read","Unexpected : "+commands[2]+", expected "+
				    type.getLocalParameterName(localRanges.size())+
				    "on line="+line_no);
			System.exit(0); 
		    }
		    if (commands[4].startsWith("range_")){
			int i=Misc.getInteger(commands[4],line_no);
			if (i>=1 && i<=ranges.size()){
			    localRanges.add(ranges.get(i-1)); 
			    // compensating for index starting at 0
			}
			else {
			    Debug.error(this,"read","Illegal integer id: "+commands[4]+
					", on line="+line_no);
			    System.exit(0); 
			}
		    }
		    else {
			Debug.error(this,"read","Unexpected range: "+commands[4]+", on line="+line_no);
			System.exit(0); 
		    }

		    
		}
		else if (commands[0].equals("has") && commands[1].equals("parameter") &&
			 commands[3].equals("=")) {
		    if (type==null) {
			Debug.error(this,"read","Reaction type must be set before "+
				    "assigning parameter, on line="+line_no);
			System.exit(0); 
		    }
		    if (!commands[2].equals(type.getLocalParameterName(localParameters.size()))){
			Debug.error(this,"read","Unexpected : "+commands[2]+", expected "+
				    type.getLocalParameterName(localParameters.size())+
				    "on line="+line_no);
			System.exit(0); 
		    }
		    if (commands.length==5){
			double d=Misc.getDouble(commands[4],line_no);
			localParameters.add(d); 
		    }
		    else {
			Debug.error(this,"read","Unexpected parameter value: , on line="+
				    line_no);
			System.exit(0); 
		    }

		    
		}
		else {
		    Debug.error(this,"read","Unexpected command: "+commands[0]+", on line="+line_no);
		    System.exit(0); 
		}
		if (localRanges.size()==type.getNumberOfLocalParameters() && 
		    localVariables.size()==type.getNumberOfLocalVariables()) break;

		if (localParameters.size()==type.getNumberOfLocalParameters() && 
		    localVariables.size()==type.getNumberOfLocalVariables()) break;
	    }
	}
	catch (IOException ioe) {
	    Debug.error(this,"read","IOExpection on line="+line_no);
	    System.exit(0); 
	}

	// Add this possible reaction to variable
	variable.addReaction(this);
	return line_no;

    }

    public ReactionType getReactionType(){
	return type;
    }

    /** Get local variable i
     * @param i the index
     * @return the variable
     */
    public Variable getLocalVariable(int i){
	return localVariables.get(i);
    }
    
    /** Get number of local variables
     * @return the number
     */ 
    public int getNumberOfLocalVariables(){
	return localVariables.size();
    }

    /** Get local range i
     * @param i the index
     * @return the range
     */ 
    public Range getLocalRange(int i){
	return localRanges.get(i);
    }
    
    /** Get number of local ranges
     * @return the number
     */ 
    public int getNumberOfLocalRanges(){
	return localRanges.size();
    }
    
    /** Get local parameter i
     * @param i the index
     * @return the parameter value
     */ 
    public double getLocalParameter(int i){
	return localParameters.get(i).doubleValue();
    } 

    /** Get number of local parameters
     * @return the number
     */ 
    public int getNumberOfLocalParameters(){
	return localParameters.size();
    }

    /** Set the reaction type
     * @param r the type
     */ 
    public  void setReactionType(ReactionType r){
	type=r;
    }
    
    /** Set the local variable i of this reaction
     * @param i the index
     * @param v the variable
     */ 
    public void setLocalVariable(int i, Variable v){
	localVariables.add(i,v);
    }
    
    /** Set the local range for parameter i of this reaction
     * @param i the index
     * @param r the range
     */ 
    public void setLocalRange(int i, Range r){
	localRanges.add(i,r);
    }

    /** Get string representation
     * @return the string
     */ 
     public String toString(){

	 String s="Reaction "+id+" "+type.getName()+"\nVariables: ";
	 for (int i=0;i<localVariables.size();i++)
	     s+=(localVariables.get(i).getId()+1)+" ";
	 s+="\n";
	 for (int i=0;i<localRanges.size();i++)
	     s+=localRanges.get(i).toString();	
	return s;
    }
  
    // public static void main(String[] argv) {
    // }
  
}
