/** 
    ODEidon, a program for structural identification of ODEs
    Copyright (C) 2011 Peter Gennemark, Dag Wedelin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Peter Gennemark, Dag Wedelin
             www.odeidentification.org

*/



package Java.problem;


import java.util.*;
import Java.debug.*;
import Java.parser.*;
import java.io.*;
import Java.misc.*;

/**
  * Description of a reaction type: <BR>
   * @author Peter Gennemark
   * @version     1.0                   
   * @since       1.0
   */
public class ReactionType  {
  
    private static int totalid=0;
    private int id;
    private String name;
    private Vector<String> localVariableNames=new Vector<String>();
    private Vector<String> localParameterNames=new Vector<String>(); 
    private Vector<String> dfdk=new Vector<String>();
    private Vector<String> dfdX=new Vector<String>();
    private String equation;
    private boolean useNumericalDerivaties=false;

    public ReactionType(){
	this.id=totalid++;
	// read database of derivatives
    }


    /** Read reaction type from input file in form of a buffered reader
     * @param commands vector of strings on first line to read
     * @param in buffered reader
     * @param line_no current line in the buffered reader (i.e., problem file)
     * @param debug debug level
     */
    public int read(String[] commands,
		    BufferedReader in,
		    int line_no,
		    int debug){
	
	id = Misc.getInteger(commands[0],line_no);
	if (id<=0) {
	    Debug.error(this,"read","id="+id+" should be positive on line="+line_no);
	    System.exit(0);
	}
	if (id!=totalid) {
	    Debug.error(this,"read","reaction_"+id+" should be reaction_"+totalid+" on line="+line_no);
	    System.exit(0);
	}
	try {
	    for (;;) {
		String s= in.readLine();
		line_no++;
		commands = Misc.split(s,line_no,debug);
		if (commands==null){
		    if (debug>0) Debug.print(this,"read","empty line");
		    continue;
		}
		else if (commands[0].equals("has") && commands[1].equals("name") &&
		    commands[2].equals("=")) name=commands[3];
		else if (commands[0].equals("has") && 
			 commands[1].startsWith("localVariableName_") && 
			 commands[2].equals("=")){
		    localVariableNames.add(commands[3]);
		}
		else if (commands[0].equals("has") && 
			 commands[1].startsWith("localParameterName_") && 
			 commands[2].equals("=")){
		    localParameterNames.add(commands[3]);
		}
		else if (commands[0].equals("has") && 
			 commands[1].equals("equation") && 
			 commands[2].equals("=")){
		    equation=commands[3];
		    break;
		}
		else {
		    Debug.error(this,"read","Unexpected command: "+commands[0]+", on line="+line_no);
		    System.exit(0); 
		}
	    }
	}
	catch (IOException ioe) {
	    Debug.error(this,"read","IOExpection on line="+line_no);
	    System.exit(0); 
	}

	readDerivatives(line_no,debug);

	return line_no;

    }

    /** Read the derivatives from file REACTIONS.txt
     * @param line_no the string
     * @param debug debug level, 0=none
     */ 
   public void readDerivatives(int line_no,
			       int debug){
       File file=new File("REACTIONS.txt");
       try {
	   BufferedReader in = new BufferedReader(new FileReader(file));
	   String line="";
	   for (;;) { 
	       if (!in.ready()) break;
	       String cline=in.readLine();
	       if (debug>0) Debug.print(this,"read",cline);
	       String[] commands = Misc.split(cline,line_no,debug);
	       if (commands==null) {
		   if (debug>0) Debug.print(this,"read","empty line");
		   continue;
	       }
	       //  ======================================================== 
	       if (commands[0].equals("name") && commands[1].equals("=") &&
		   commands[2].equals(name)) {
		   if (debug>0) Debug.print(this,"readDerivatives","Found type in database");


		   String s= in.readLine();
		   commands = Misc.split(s,line_no,debug);
		   if (commands[0].equals("equation") && 
		       commands[1].equals("=") && 
		       commands[2].equals(equation)) {
		       
		       for (int i=0;i<localParameterNames.size();i++) {
			   s= in.readLine();
			   commands = Misc.split(s,line_no,debug);
			   if (commands==null || commands.length!=3){
			       Debug.error(this,"readDerivatives",
					   "In REACTIONS.txt: found: "+s+
					   ", expected: dfdk"+(i+1)+" = ....");
			       System.exit(0);
			   }
			   dfdk.add(commands[2]);
		       }
		       for (int i=0;i<localVariableNames.size();i++) {
			   s= in.readLine();
			   commands = Misc.split(s,line_no,debug);
			   if (commands==null || commands.length!=3){
			       Debug.error(this,"readDerivatives",
					   "In REACTIONS.txt: found: "+s+
					   ", expected: dfdX"+(i+1)+" = ....");
			       System.exit(0);
			   }
			   dfdX.add(commands[2]);
		       }
		   } 
		   else {
		       Debug.error(this,"readDerivatives",
				   "Found: equation = "+equation+
				   "; expected: "+commands[0]+" "+
				   commands[1]+" "+commands[2]);
		       System.exit(0);
		   }
	       }
	   }

	   if (dfdX.size()!=localVariableNames.size() ||
	       dfdk.size()!=localParameterNames.size()) {
	       Debug.warning(this,"readDerivatives","Reaction type "+
			     name+" did not found in "+file);
	       Debug.warning(this,"readDerivatives",
			     "Consider choosing one of the predefined reaction types in "+file);
	       Debug.warning(this,"readDerivatives","or complement "+
			     file+" with your own new reaction type.");
	       useNumericalDerivaties=true;
	   }

       }catch(IOException ioe) {    
	   String errorMsg="Error reading file " + file;   
	   Debug.print(this,"readDerivatives",errorMsg); 
	   //System.exit(0);	
	   javax.swing.JOptionPane.showMessageDialog(null,
						     errorMsg,
						     "Error" ,
						     javax.swing.JOptionPane.ERROR_MESSAGE); 
       } 
   }
    

    /** Set as a GMA term with gma variables,
     * e.g., alpha*X1^g11*X2^g12*X3^g13 for gma=3. 
     * @param gma the number of variables
     */
    public void setAsGMAterm(int gma){
	id = gma+1;
	name="GMAterm"+gma+"Variable";
	if (gma!=1) name+="s";
	for (int i=1;i<=gma;i++) localVariableNames.add("X"+i);
	for (int i=1;i<=gma+1;i++) localParameterNames.add("k"+i);
	equation="k1";
	for (int i=1;i<=gma;i++) equation+="*X"+i+"^k"+(i+1);
	readDerivatives(0,0);
    }

    /** Get the name of local variable i in the kinetic equation
     * @param i variable index
     * @return the name
     */
    public String getLocalVariableName(int i){
	return localVariableNames.get(i);
    } 

    /** Get the name of local parameter i in the kinetic equation
     * @param i parameter index
     * @return the name
     */
    public String getLocalParameterName(int i){
	return localParameterNames.get(i);
    }
    
    /** Get the number of local variables in the kinetic equation
     * @return the number
     */
    public int getNumberOfLocalVariables(){
	return localVariableNames.size();
    }
    
    /** Get the number of local parameters in the kinetic equation
     * @return the number
     */
    public int getNumberOfLocalParameters(){
	return localParameterNames.size();
    }

    /** Get the partial derivative of the kinetic equation wrt variable ivar
     * @param ivar the variable index
     * @return the partial derivative, null if not defined
     */ 
    public String get_dfdX(int ivar){
	if (ivar>=dfdk.size()) return null;
	return dfdX.get(ivar);
    }
    
    /** Get the partial derivative  of the kinetic equation wrt parameter ipara
     * @param ipara parameter index
     * @return the partial derivative, null if not defined
     */ 
    public String get_dfdk(int ipara){
	if (ipara>=dfdk.size()) return null;
	return dfdk.get(ipara);
    }

    /** Monitor if analytical derivatives are defined.
     * @return true if numerical derivatives will be used, otherwise analytical derivatives will be used
     */ 
    public boolean useNumericalDerivaties(){
	return useNumericalDerivaties;
    }

    /** Get the name of the reaction type
     * @return the name
     */ 
    public String getName(){
	return name;
    }

    /** Get the kinetic equation
     * @return the kinetic equation
     */ 
    public String getEquation(){
	return equation;
    }

    /** Get SBML representation of the reaction equation
     * @return the SBML representation
     */ 
    public String getSBML(){
	FunctionParser fp=new FunctionParser(equation);
	return fp.getSBML();
    }

    /** Get id of the reaction type
     * @return the id
     */ 
    public int getId(){ 
	return id;
    }


    /** Get string representation
     * @return the string
     */ 
    public String toString(){
	String s="Reaction type "+id+" "+name+"\n";

	s+="Variables: ";
	for (int i=0;i<getNumberOfLocalVariables();i++)
	    s+=localVariableNames.get(i)+" ";
	s+="\n";

	s+="Parameters: ";
	for (int i=0;i<getNumberOfLocalParameters();i++)
	    s+=localParameterNames.get(i)+" ";
	s+="\n";

	s+="equation: "+equation+"\n";
	s+="dfdk: ";
	for (int i=0;i<dfdk.size();i++)
	    s+=dfdk.get(i)+" ";
	s+="\n";
	s+="dfdX: ";
	for (int i=0;i<dfdX.size();i++)
	    s+=dfdX.get(i)+" ";
	s+="\n";
	return s;
    }

  
    // public static void main(String[] argv) {
    // }
  
}
