/** 
    ODEidon, a program for structural identification of ODEs
    Copyright (C) 2011 Peter Gennemark, Dag Wedelin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Peter Gennemark, Dag Wedelin
    www.odeidentification.org

*/



package Java.problem;

import java.util.*;
import Java.debug.*;
import java.io.*;
import Java.misc.*;

/**
 * Description of a sample: 
 * <BR> Vector of data-points (values) and standard deviations (sdevs) for each variable at time-point time.
 * <BR> isAvailable is false if data-point is missing.
 * @author Peter Gennemark
 * @version     1.0                   
 * @since       1.0
 */


public class Sample {

    private static int totalid=0;
    private int id;
    private double time=0.0; 
    private Vector<Double> values; 
    private Vector<Double> sdevs; 
    private Vector<Boolean> isAvailable=new Vector<Boolean>(); 
   
    public Sample(){
	this.id=totalid++;
    }
   
    /** Read sample from input file in form of a buffered reader
     * @param commands vector of strings on first line to read
     * @param in buffered reader
     * @param line_no current line in the buffered reader (i.e., problem file)
     * @param debug debug level
     * @param variables vector of variables
     * @param experiments vector of experiments
     */
    public int read(String[] commands,
		    BufferedReader in,
		    int line_no,
		    int debug,
		    Vector<Variable> variables,
		    Vector<Experiment> experiments){
	
	boolean timeIsRead=false;
	id = Misc.getInteger(commands[0],line_no);
	if (id<=0) {
	    Debug.error(this,"read","id="+id+" should be positive on line="+line_no);
	    System.exit(0);
	}
	if (id!=totalid && id==1) {
	    totalid=1; // start a new experiment
	}
	if (id!=totalid) {
	    Debug.error(this,"read","sample_"+id+" should be sample_"+totalid+" on line="+line_no);
	    System.exit(0);
	}
	Experiment experiment=null;
	
	if (commands[1].equals("of") && commands[2].startsWith("experiment_")){
	    int iexp=Misc.getInteger(commands[2],line_no);
	    if (iexp>=1 && iexp<=experiments.size()){
		experiment=experiments.get(iexp-1); // compensating for index starting at 0
	    }
	    else{
		Debug.error(this,"read","Illegal experiment index: "+commands[2]+", on line="+line_no);
		System.exit(0); 
	    }
	}
	else {
	    Debug.error(this,"read","Unexpected command: "+commands[1]+" "+
			commands[2]+", on line="+line_no);
	    System.exit(0); 
	}

    
	try {
	    for (;;) {
		String s= in.readLine();
		line_no++;
		commands = Misc.split(s,line_no,debug);
		if (commands==null){
		    if (debug>0) Debug.print(this,"read","empty line");
		    continue;
		}
		else if (commands[0].equals("has") && commands[1].equals("time") &&
			 commands[2].equals("=")) {
		    time=Misc.getDouble(commands[3],line_no);
		    timeIsRead=true;
		}
		else if (commands[0].equals("has") && commands[1].equals("variable_") &&
			 commands[2].equals("=")) {
		    if (commands.length-3!=variables.size()){
			Debug.error(this,"read","Number of data="+(commands.length-3)+
				    " is not equal to number of variables="+variables.size()+
				    ", on line="+line_no);
			System.exit(0); 
		    }
		    values=new Vector<Double>();
		    for (int i=0;i<variables.size();i++){
			if (commands[3+i].equals("NA")){
			    values.add(new Double(0.0));
			    isAvailable.add(new Boolean(false));
			}
			else {
			    values.add(Misc.getDouble(commands[3+i],line_no));
			    isAvailable.add(new Boolean(true));
			}
		    }

		}	
		else if (commands[0].equals("has") &&commands[1].equals("sdev") &&
			 commands[2].equals("of") && commands[3].equals("variable_") &&
			 commands[4].equals("=")) {
		    if (commands.length-5!=variables.size()){
			Debug.error(this,"read","Number of data="+(commands.length-5)+
				    " is not equal to number of variables="+variables.size()+
				    ", on line="+line_no);
			System.exit(0); 
		    }
		    sdevs=new Vector<Double>();
		    for (int i=0;i<variables.size();i++){
			if (isAvailable.get(i)) {
			    double sd=Misc.getDouble(commands[5+i],line_no);
			    if (sd<=0.0 && variables.get(i).isDependent()){
				Debug.warning(this,"read","Sdev="+sd+" should be positive"+
					      ", on line="+line_no);
			    }
			    sdevs.add(sd);
			}
			else {
			    sdevs.add(new Double(1.0));
			}
		    }
		    
		}
		else {
		    Debug.error(this,"read","Unexpected command: "+s+", on line="+line_no);
		    System.exit(0); 
		}
		if (values!=null  && sdevs!=null && timeIsRead) break;
	    }
	}
	catch (IOException ioe) {
	    Debug.error(this,"read","IOExpection on line="+line_no);
	    System.exit(0); 
	}

	// Add this possible reaction to variable
	experiment.addSample(this);
	return line_no;

    }

    /** Get string representation
     * @return the string
     */ 
    public String toString() {
	String s="Sample "+id+", time="+time+"\n y:    ";

	for (int i=0;i<values.size();i++)
	    s+=values.get(i)+" ";
	s+="\n sdev: ";
	for (int i=0;i<sdevs.size();i++)
	    s+=sdevs.get(i)+" ";
	s+="\n";

	return s;
    }
    
    /** Get time-point of the sample
     * @return the time-point
     */ 
    public double getTime(){return time;}

    /** Get value of the sample for variable i
     * @return the value for variable i
     */ 
    public double getValue(int ivar){
	return values.get(ivar); // value for variable ivar 0...nvar-1
    }

    /** Get standard deviation of the sample for variable i
     * @return the standard deviation for variable i
     */
    public double getSdev(int ivar){
	return sdevs.get(ivar); // sdev for variable ivar 0...nvar-1
    }
     
    // public static void main(String[] argv) {
    // }
    
}
