/** 
    ODEidon, a program for structural identification of ODEs
    Copyright (C) 2011 Peter Gennemark, Dag Wedelin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Peter Gennemark, Dag Wedelin
             www.odeidentification.org

*/


package Java.problem;

import java.util.*;
import Java.debug.*;
import java.io.*;
import Java.misc.*;

/**
  * Description of a variable
  * @author Peter Gennemark
  * @version     1.0                   
  * @since       1.0
  */
public class Variable  {
    private static int totalid=0;
    private int id;
    private String name;
    private String type;
    private Vector<PossibleReaction> possibleReactions=new Vector<PossibleReaction>();
    private Vector<Reaction> reactions=new Vector<Reaction>(); // initial model
    
    public Variable(){
	this.id=totalid;
	totalid++;
    }


    /** Read variable from input file in form of a buffered reader
     * @param commands vector of strings on first line to read
     * @param in buffered reader
     * @param line_no current line in the buffered reader (i.e., problem file)
     * @param debug debug level
     */
    public int read(String[] commands,
		    BufferedReader in,
		    int line_no,
		    int debug){
	
	id = Misc.getInteger(commands[0],line_no)-1;
	if (id<0) {
	    Debug.error(this,"read","id="+id+" should be positive on line="+line_no);
	    System.exit(0);
	}
	if (id!=totalid-1) {
	    Debug.error(this,"read","variable_"+id+" should be variable_"+(totalid-1)+" on line="+line_no);
	    System.exit(0);
	}
	if (commands[1].equals("has") && commands[2].equals("name") &&
	    commands[3].equals("=")) {
	    name=commands[4];
	}
	else {
	    Debug.error(this,"read","Unexpected command: "+commands[0]+", on line="+line_no);
	    System.exit(0); 
	}
	if (commands[5].equals("is") && (commands[6].equals("dependent") || 
					 commands[6].equals("inputVariable"))){
		type=commands[6];
	}
	else {
	    Debug.error(this,"read","Unexpected command: "+commands[0]+", on line="+line_no);
	    System.exit(0); 
	}
   	return line_no;
	
    }

    /** Check if type of variable is inputVariable
     * @return true if inputVariable, ow false
     */
    public boolean isInputVariable(){
	return type.equals("inputVariable");
    }

    /** Check if type of variable is dependent
     * @return true if dependent, ow false
     */
    public boolean isDependent(){
	return type.equals("dependent");
    }

    /** Get type of variable; dependent or inputVariable
     * @return the type
     */
    public String getType(){
	return type;
    }

    /** Get id of variable, numbered 0...N-1
     * @return the id
     */ 
    public int getId(){
	return id;
    }

    /** Get name of variable
     * @return the name
     */ 
    public String getName(){
	return name;
    }

   /** Get possible reaction i assigned to the model space of this variable
     * @param i the index
     * @return the possible reaction i
     */
    public PossibleReaction getPossibleReaction(int i){
	return possibleReactions.get(i);
    }
    
    /** Get number of possible reactions assigned to the 
     * model space of this variable
     * @return the number of reactions
     */
    public int getNumberOfPossibleReactions(){
	return possibleReactions.size();
    }
    
    /** Add possible reaction r to the model space of this variable
     * @param r the reaction to add
     */ 
    public void addPossibleReaction(PossibleReaction r){
	possibleReactions.add(r);
    }

    /** Get reaction i assigned to this variable
     * @param i the index
     * @return the reaction i
     */
    public Reaction getReaction(int i){
	return reactions.get(i);
    }
    
    /** Get number of reactions assigned to this variable
     * @return the number of reactions
     */
    public int getNumberOfReactions(){
	return reactions.size();
    }
    
    /** Add reaction r to this variable
     * @param r the reaction to add
     */ 
    public void addReaction(Reaction r){
	reactions.add(r);
    }
    
    /** Set list of reactions to the empty list
     */ 
    public void emptyReactions(){
	reactions=new Vector<Reaction>(); 
    }
  

    /** Get string representation
     * @return the string
     */ 
    public String toString(){
	String s="Variable "+id+" name: "+name+" type: "+type+"\n";

	for (int i=0;i<getNumberOfPossibleReactions();i++)
	    s+=possibleReactions.get(i)+" ";
	s+="\nReactions:\n";
	for (int i=0;i<getNumberOfReactions();i++)
	    s+=reactions.get(i)+" ";
	s+="\n";
	
	return s;
    }


    //public static void main(String[] argv) {
    //}
  
}
