/** 
    ODEidon, a program for structural identification of ODEs
    Copyright (C) 2011 Peter Gennemark, Dag Wedelin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Peter Gennemark, Dag Wedelin
             www.odeidentification.org

*/



package Java.problem;

import java.util.*;
import Java.debug.*;
import java.io.*;
import Java.misc.*;

/**
  * Description of a variable set <BR>
  * @author Peter Gennemark
  * @version     1.0                   
  * @since       1.0
  */
public class VariableSet  {
    private static int totalid=0;
    private int id;
    private String name;
    private Vector<Variable> members=new Vector<Variable>();

    
    public VariableSet(){
	this.id=totalid++;
    }

    /** Create variable set containing one variable
     * @param v add one single member
     */
    public VariableSet(Variable v){
	name="Temporary";
	id=0;
	members.add(v);
    }


    /** Create variable set containing a vector of variables
     * @param members vector of variables
     */
    public VariableSet(Vector<Variable> members){
	name="Temporary";
	id=0;
	this.members=members;
    }


    /** Read variable set form input file in form of a buffered reader
     * @param commands vector of strings on first line to read
     * @param in buffered reader
     * @param line_no current line in the buffered reader (i.e., problem file)
     * @param debug debug level
     * @param variables vector of variables
     */
    public int read(String[] commands, 
		    BufferedReader in, 
		    int line_no, 
		    int debug,
		    Vector<Variable> variables){
	id = Misc.getInteger(commands[0],line_no);
	if (id<=0) {
	    Debug.error(this,"read","id="+id+
			" should be positive on line="+line_no);
	    System.exit(0);
	}
	if (id!=totalid) {
	    Debug.error(this,"read","memberOfVarSet_"+id+
			" should be memberOfVarSet_"+totalid+
			" on line="+line_no);
	    System.exit(0);
	}
	for (int i=1;i<commands.length;i++){
	    if (commands[i].startsWith("variable_")) {
		int ivar=Misc.getInteger(commands[i],line_no);
		if (ivar>=1 && ivar<=variables.size()){
		    members.add(variables.get(ivar-1)); // compensating for index starting at 0
		}
	    }
	    else {
		Debug.error(this,"read","Unexpected command: "+commands[i]+", on line="+line_no);
		System.exit(0); 
	    }
	}
	return line_no;
	    
    }

   /** Get the number of member variables
     * @return the number of member variables
     */
    public int getNumberOfMembers(){
	return members.size();
    }
    
    /** Get member variable i from the set of variables
     * @return variable i
     */
    public Variable getMember(int i){
	if (i<0 || i>=members.size()) {
	    Debug.error(this,"getMember","VariableSet.getMember: input "+i+
			" out of legal range");
	    return  null;
	}
	return members.get(i);
    }
       
    /** Get string representation
     * @return the string
     */ 
    public String toString(){
	String s="VariableSet "+id+". ";
	s+="Members: ";
	for (int i=0;i<getNumberOfMembers();i++)
	    s+=members.get(i).getId()+" ";
	s+="\n";
	return s;
    }
    
    // public static void main(String[] argv) {
    // }
  
}
