/** 
    ODEidon, a program for structural identification of ODEs
    Copyright (C) 2011 Peter Gennemark, Dag Wedelin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Peter Gennemark, Dag Wedelin
    www.odeidentification.org

*/



package Java.solution;

import java.io.*;
import java.util.*;
import Java.debug.*;
import java.lang.Integer;
import Java.problem.*;
import Java.misc.*;

/**
 * Description of a solution to a problem<BR>
 * @author Peter Gennemark
 * @version     1.0                   
 * @since       1.0
 */
public class Solution {
    
        
    private Problem problem;
    private String name;
    private String type;
    private String date;
    private Vector<ReactionType> reactionTypes;
    private Vector<Variable> variables;
    private double error;
    private double residual;
    private int[] RAND=new int[2];
    private double smoothingParameter;
    private int SEED;
    
    
    /** To read a solution requires reactionsTypes and variables from the problem
     *  @param p the problem from which the solution has been generated
     */
    public Solution(Problem p) {
	this.reactionTypes=p.getReactionTypes();
	this.variables=p.getVariables();
	this.problem=p;
	// Initially we remove all reactions from each variable
	for (int i=0;i<p.getNumberOfVariables();i++){ 
	    variables.get(i).emptyReactions();
	}
    }
        
    
    /** Read the solution file
     * @param file to be read
     * @param debug debug level; 0=none
     */
    public void read(File file,int debug) {
	
	int line_no=0;
	Debug.print(this,"read","File: " + file);
	try {
	    BufferedReader in = new BufferedReader(new FileReader(file));
	    String line="";
       
	    for (;;) { 
	  
		if (!in.ready()) break;
		String cline=in.readLine();
		line_no++;
		if (debug>0) Debug.print(this,"read",cline);
		String[] commands = Misc.split(cline,line_no,debug);
		if (commands==null) {
		    if (debug>0) Debug.print(this,"read","empty line");
		    continue;
		}
	  
		//  ======================================================== // comment
		if (commands[0].startsWith("//")) {
		    if (debug>0) Debug.print(this,"read","comment");
		}
	 
		//  ======================================================== format
		else if (commands[0].equals("format")) {
		    if (debug>0) Debug.print(this,"read","format");
		}
	 
		//  ======================================================== begin
		else if (commands[0].equals("begin")) {
		    if (debug>0) Debug.print(this,"read","begin");
		    if (commands.length<5){
			Debug.error(this,"read","expected 'begin solution of problem...'"+
				    " found only "+cline+", on line="+line_no);
			System.exit(0);
		    }
		    if (commands[1].equals("solution") && 
			commands[2].equals("of") && 
			commands[3].equals("problem")) {
			name=commands[4];
		    }
		    else{
			Debug.error(this,"read","expected 'solution of problem' found "+
				    commands[1]+
				    commands[2]+
				    commands[3]+
				    " on line="+line_no);
			System.exit(0);
		    }
		}

		//  ======================================================== date
		else if (commands[0].equals("date")) {
		    if (commands.length==4 && 
			commands[1].equals("=")) {
			date=commands[2]+" "+commands[3];
		    }
		    else {
			Debug.error(this,"read","expected 'date = YYYYMMDD HHMM' found "+
				    cline+
				    " on line="+line_no);
			System.exit(0);
		    }
		    if (debug>0) Debug.print(this,"read","date="+date);

		}

		//  ======================================================== url 
		else if (commands[0].equals("url")) {
		    if (debug>0) Debug.print(this,"read","url");
		}

		//  ======================================================== randomStarts_
		else if (commands[0].equals("randomStarts_")) {
		    if (debug>0) Debug.print(this,"read","randomStarts_"); 
		    if (commands[1].equals("=")) {
			RAND[0]= Misc.getInteger(commands[2],line_no);
			RAND[1]= Misc.getInteger(commands[3],line_no);
		    }
		    else{
			Debug.error(this,"read","expected = found "+commands[1]+" on line="+line_no);
			System.exit(0);
		    }
		}
	 
		//  ======================================================== smoothingParameter
		else if (commands[0].equals("smoothingParameter")) {
		    if (debug>0) Debug.print(this,"read","smoothingParameter"); 
		    if (commands[1].equals("=")) {
			smoothingParameter=Misc.getDouble(commands[2],line_no);
		    }
		    else{
			Debug.error(this,"read","expected = found "+commands[1]+" on line="+line_no);
			System.exit(0);
		    }
		}
	    
		//  ======================================================== randomSeed
		else if (commands[0].equals("randomSeed")) {
		    if (debug>0) Debug.print(this,"read","randomSeed"); 
		    if (commands[1].equals("=")) {
			SEED=Misc.getInteger(commands[2],line_no);
		    }
		    else{
			Debug.error(this,"read","expected = found "+commands[1]+" on line="+line_no);
			System.exit(0);
		    }
		}

		//  ======================================================== reaction_
		else if (commands[0].startsWith("reaction_")) {
		    if (debug>0) Debug.print(this,"read","reaction_");


		    if (commands.length==3 && 
			commands[1].equals("of") && 
			commands[2].startsWith("variable_")) {
			Reaction r=new Reaction();
			line_no=r.read(commands,in,line_no,debug,variables,
				       reactionTypes,null); //also adds to variable, ranges is null
			if (debug>0) Debug.print(this,"read",r.toString());
		    }
		    else {
			Debug.error(this,"read","expected 'reaction_i of variable_j' found "+
				    cline+
				    " on line="+line_no);
			System.exit(0);
		    }
		}

		//  ======================================================== error
		else if (commands[0].equals("error")) {
		    if (commands.length==3 && commands[1].equals("="))
			error=Misc.getDouble(commands[2],line_no);
		    if (debug>0) Debug.print(this,"read","error="+error);
		    else {
			Debug.error(this,"read","expected 'error = real_number' found "+
				    cline+", on line="+line_no);
			System.exit(0);
		    }
		}

		//  ======================================================== residual
		else if (commands[0].equals("residual")) {
		    if (commands.length==3 && commands[1].equals("="))
			residual=Misc.getDouble(commands[2],line_no);
		    if (debug>0) Debug.print(this,"read","residual="+residual);
		    else {
			Debug.error(this,"read","expected 'residual = real_number' found "+
				    cline+", on line="+line_no);
			System.exit(0);
		    }
		}

		//  ======================================================== end
		else if (commands[0].equals("end")) {
		    if (debug>0) Debug.print(this,"read","end");
		    //break; // don not break here, random seed may be defined manually after end
		}

		else{
		    Debug.error(this,"read","Illegal command: "+cline+", line: " + line_no);
		    System.exit(0);
		}

	    } //main loop



	}catch(IOException ioe) {    
	    String errorMsg="Error reading file " + file + ", line number: " + line_no;   
	    Debug.print(this,"read",errorMsg); 
	    //System.exit(0);	
	} 

    }//end of read


    /** save the solution model as SBML
     * @param file to save in
     * @param solname name of solution
     * @param debug debug level; 0=none
     */
    public void saveSBML(File file,String solname, int debug) {
	try {
	    BufferedWriter p = new BufferedWriter(new FileWriter(file));
	    
	    p.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	    p.write("<sbml xmlns=\"http://www.sbml.org/sbml/level2\" level=\"2\" version=\"1\">\n");
	    p.write("<model id=\""+solname+"\" name=\""+solname+"\">\n");
	    p.write("<notes>\n");
	    p.write("<html xmlns=\"http://www.w3.org/1999/xhtml\">\n");
	    p.write("<body>\n");
	    p.write("<p>This model is a solution to identification problem "+name+",</p>\n");
	    p.write("<p>see <a href=\"www.odeidentification.org\">\n");
	    p.write("Benchmarks for identification of ODEs</a>.</p>\n");
	    p.write("<p>If several experimental conditions are\n");
	    p.write(" considered, the first is given here.</p>\n");
	    p.write("<p>Model created "+date+"</p>\n"); 
	    Calendar cal = Calendar.getInstance();
	    java.text.SimpleDateFormat sdf = new java.text.SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
	    p.write("<p>This file created "+sdf.format(cal.getTime())+"</p>\n");      
	    p.write("</body>\n"); 
	    p.write("</html>\n"); 
	    p.write("</notes>\n");
	    p.write("<listOfCompartments>\n");
	    p.write("<compartment id=\"default\" size=\"1\"/>\n");
	    p.write("</listOfCompartments>\n");
	    p.write("<listOfSpecies>\n");
	    for (int i=0;i<variables.size();i++){
		Variable v=variables.get(i);
		p.write("<species id=\""+v.getName()+"\""+
			" name=\""+v.getName()+"\""+
			" compartment=\"default\""+
			" initialAmount=\""+problem.getExperiment(0).getSample(0).getValue(i)+"\""+
			" charge =\"0\"/>\n");
	    }
	    p.write("</listOfSpecies>\n");
	    p.write("<listOfReactions>\n");
	    for (int ivar=0;ivar<variables.size();ivar++){
		Variable v=variables.get(ivar);
		if (v.isDependent()) {
		    for (int ireact=0;ireact<v.getNumberOfReactions();ireact++){
			Reaction r=v.getReaction(ireact);
			p.write("<reaction id=\"re"+(ivar+1)+(ireact+1)+"\""+
				" reversible=\"false\" fast=\"false\">\n");
			p.write("<listOfProducts>\n");
			p.write("<speciesReference species=\""+v.getName()+"\"/>\n");
			p.write("</listOfProducts>\n");

			// Modifier
			if (r.getNumberOfLocalVariables()==0){
			    // no modifier
			}
			else if (r.getNumberOfLocalVariables()==1 && r.getLocalVariable(0).getId()==v.getId()){
			    // no modifier
			}
			else { // at least one modifier
			    p.write("<listOfModifiers>\n");
			    for (int iloc=0;iloc<r.getNumberOfLocalVariables();iloc++){
				Variable vloc=r.getLocalVariable(iloc);
				if (v.getId()!=vloc.getId()) {
				    p.write("<modifierSpeciesReference species=\""+
					    vloc.getName()+"\"/>\n");
				}
			    }
			    p.write("</listOfModifiers>\n"); 
			}

			// Kinetic law
			String sbml=r.getReactionType().getSBML();


			for (int iloc=0;iloc<r.getNumberOfLocalParameters();iloc++){
			    sbml=sbml.replace("k"+(iloc+1),"Para"+(ivar+1)+(ireact+1)+(iloc+1));
			    // use temp name Para, ow several changes may occur to one parameter
			}
			sbml=sbml.replace("Para","k");
			for (int iv=0;iv<r.getNumberOfLocalVariables();iv++){
			    int id=r.getLocalVariable(iv).getId();
			    sbml=sbml.replace("X"+(iv+1),"x"+(id+1));
			}
			p.write("<kineticLaw>\n");
			p.write("<math xmlns=\"http://www.w3.org/1998/Math/MathML\">\n");
			p.write(sbml);
			p.write("</math>\n");

			// Parameter
			if (r.getNumberOfLocalParameters()==0){
			    // no parameter
			}
			else { // at least one parameter
			    p.write("<listOfParameters>\n");
			    for (int iloc=0;iloc<r.getNumberOfLocalParameters();iloc++){
				double ploc=r.getLocalParameter(iloc);
				p.write("<parameter id=\"k"+(ivar+1)+(ireact+1)+(iloc+1)+"\""+
					" name=\"k"+(ivar+1)+(ireact+1)+(iloc+1)+"\""+
					" value=\""+ploc+"\"/>\n");
			    }
			    p.write("</listOfParameters>\n");
			}
			p.write("</kineticLaw>\n");
			p.write("</reaction>\n");
		    } // end for
		}// endif
	    } // end loop
	    p.write("</listOfReactions>\n");
	    p.write("</model>\n");
	    p.write("</sbml>\n"); 
	    p.close();
	}catch(IOException ioe) {    
	    Debug.print(this,"saveSBML",file+""); 
	} 
    }
    

    /** String value of solution
     * @return the string value of the solution
     */  
    public String toString(){
	String ret="";
	ret+=name+" error="+error+"\n";
	return ret;
    }
    
    /** Get random seed used to generate solution
     * @return the seed
     */ 
    public int getSEED(){
	return SEED;
    } 
    
    /** Get the smoothing parameter used to process data in problem
     * @return the smoothing parameter
     */
    public double getSmoothingParameter(){
	return smoothingParameter;
    }
    
    /** Get the number of random starts in local parameter estimation
     * @return the number of random starts
     */
    public int getRAND1(){
	return RAND[0];
    }

    /** Get the number of random starts in global parameter estimation
     * @return the number of random starts
     */
    public int getRAND2(){
	return RAND[1];
    }

    /** Get the name of the solution
     * @return the name
     */
    public String getName(){
	return name;
    }




    // public static void main(String[] argv){  
    //   // File system variables
    //   String filetree = System.getProperty("user.dir") + 
    //     System.getProperty("file.separator");
    //   // Problem file
    //   File file=new File(filetree+argv[0]);
    //   Problem p=new Problem();
    //   p.read(file,0);
    //   // Solution file
    //   File file2=new File(filetree+argv[1]);
    //   Solution s=new Solution(p);
    //   s.read(file2,1);    
    //   if (argv.length>2){
    // 	File file3=new File(filetree+argv[2]);
    // 	s.saveSBML(file3,file2.getName(),1);
    //   }
    // }

    
}


